/* trspwhash
 * Usage:  trspwhash password	  // Hash a password
 *         trspwhash -u 0xhash    // Unhash a password to letters
 *         trspwhash -n 0xhash    // Unhash a password to letters and digits
 */
#include <stdio.h>

unsigned int
pwhash(unsigned char pw[8])
{
    unsigned char *p = &pw[7];
    unsigned int count = 8;
    unsigned int hl, t1, t2;

    hl = 0xffff;
    do {
	t1 = hl & 0x07;
	t2 = hl & 0xff;
	hl = (t1 << 13) ^ (t1 << 9) ^ (t1 << 2) ^
	     (t2 << 8) ^ (t2 << 4)  ^ (t2 >> 3) ^
             (hl >> 8) ^ (*p-- << 8);
    } while (--count);
    return hl;
}

void
usage()
{
    fprintf(stderr, "usage: trspwhash [-u | -n] arg\n");
    exit(1);
}

int
main(int argc, char **argv)
{
    unsigned int goal;
    unsigned char pw[16];
    int i;
    if (argc == 2) {
	strncpy(pw, argv[1], 8);
	pw[8] = '\0';
	strcat(pw, "        ");
	for (i = 0; i < 8; i++) {
	    if (islower(pw[i])) pw[i] = toupper(pw[i]);
	}
	printf("%04x\n", pwhash(pw));
    } else if (argc == 3 && strcmp(argv[1], "-u") == 0) {
	goal = strtoul(argv[2], (void*)0, 0);
	strcpy(pw, "        ");
	for (;;) {
	    if (pwhash(pw) == goal) printf("%s\n", pw);
	    i = 0;
	    for (;;) {
		switch (pw[i]) {
		  case ' ':
		    pw[i] = 'A';
		    break;
		  case 'Z':
		    pw[i] = 'A';
		    i++;
		    if (i == 8) exit(0);
		    continue;
		  default:
		    pw[i]++;
		    break;
		}
		break;
	    }
	}
    } else if (argc == 3 && strcmp(argv[1], "-n") == 0) {
	goal = strtoul(argv[2], (void*)0, 0);
	strcpy(pw, "        ");
	for (;;) {
	    if (pwhash(pw) == goal) printf("%s\n", pw);
	    i = 0;
	    for (;;) {
		switch (pw[i]) {
		  case ' ':
		    pw[i] = 'A';
		    break;
		  case 'Z':
		    pw[i] = '0';
		    break;
		  case '9':
		    pw[i] = 'A';
		    i++;
		    if (i == 8) exit(0);
		    continue;
		  default:
		    pw[i]++;
		    break;
		}
		break;
	    }
	}		
    } else {
	usage();
    }
    return 0;
}
