;---------------------------------------------------------------------
; TRSDOS 6.x Operating System Interface          (C) 2009 Miguel Dutra
;---------------------------------------------------------------------

;---------------------------------------------------------------------
; TRSDOS GAT disk configuration flags
;---------------------------------------------------------------------

TD_GAT_Type                 equ     10000000y                           ; 0:System Disk, 1:Data Disk
TD_GAT_Density              equ     01000000y                           ; 0:Single Density, 1:Double Density
TD_GAT_Sides                equ     00100000y                           ; 0:Single Sided, 1:Double Sided
TD_GAT_Granules             equ     00000111y                           ; Number of granules per track minus 1

;---------------------------------------------------------------------
; TRSDOS File Primary Directory Entry attributes
;---------------------------------------------------------------------

TD_FPDE_Extended            equ     10000000y                           ; 0:File Primary Directory Entry, 1:File Extended Directory Entry
TD_FPDE_System              equ     01000000y                           ; 0:Normal file, 1:System file
TD_FPDE_PDS                 equ     00100000y                           ; 1:Partition Data Set (PDS) file
TD_FPDE_Active              equ     00010000y                           ; 0:Directory Record inactive, 1:Directory Record in use
TD_FPDE_Invisible           equ     00001000y                           ; 0:File Visible, 1:File Invisible
TD_FPDE_Access              equ     00000111y                           ; USER protection level (0=Full, 1=Remove, 2=Rename, 3=Write, 4=Update, 5=Read, 6=Execute, 7=No Access)

TD_FPDE_Created             equ     10000000y                           ; 1:File was CREATEd
TD_FPDE_Modified            equ     01000000y                           ; 1:Backup pending
TD_FPDE_Open                equ     00100000y                           ; 1:File in an open condition with UPDATE access or greater
TD_FPDE_BadDate             equ     00010000y                           ; 1:The date of the next bits is inaccurate
TD_FPDE_Month               equ     00001111y                           ; Binary month of the last modification date

TD_FPDE_Day                 equ     11111000y                           ; Binary day of last modification
TD_FPDE_Year                equ     00000111y                           ; Binary year minus 80 (1980 = 000, 1981 = 001...)

;---------------------------------------------------------------------
; TRSDOS function call offsets
;---------------------------------------------------------------------

TD_Call_Load                equ     0
TD_Call_Unload              equ     4
TD_Call_GetDosVersion       equ     8
TD_Call_GetDiskName         equ     12
TD_Call_SetDiskName         equ     16
TD_Call_GetDiskDate         equ     20
TD_Call_SetDiskDate         equ     24
TD_Call_FindFiles           equ     28
TD_Call_GetFileHandle       equ     32
TD_Call_GetFileName         equ     36
TD_Call_GetFileExtension    equ     40
TD_Call_GetFileSize         equ     44
TD_Call_GetFileDate         equ     48
TD_Call_SetFileDate         equ     52
TD_Call_GetFileAttributes   equ     56
TD_Call_SetFileAttributes   equ     60
TD_Call_GetFilePointer      equ     64
TD_Call_SetFilePointer      equ     68
TD_Call_ReadFile            equ     72
TD_Call_WriteFile           equ     76
TD_Call_RenameFile          equ     80
TD_Call_DeleteFile          equ     84
TD_Call_Hash                equ     88
TD_Call_DEC2FDE             equ     92
TD_Call_Sector2CHS          equ     96

.data

;---------------------------------------------------------------------
; TRSDOS Virtual Table
;---------------------------------------------------------------------

TRSDOS_VTBL             DWORD   TD_Load                                 ; OSI standard member functions
                        DWORD   TD_Unload
                        DWORD   TD_GetDosVersion
                        DWORD   TD_GetDiskName
                        DWORD   TD_SetDiskName
                        DWORD   TD_GetDiskDate
                        DWORD   TD_SetDiskDate
                        DWORD   TD_FindFiles
                        DWORD   TD_GetFileHandle
                        DWORD   TD_GetFileName
                        DWORD   TD_GetFileExtension
                        DWORD   TD_GetFileSize
                        DWORD   TD_GetFileDate
                        DWORD   TD_SetFileDate
                        DWORD   TD_GetFileAttributes
                        DWORD   TD_SetFileAttributes
                        DWORD   TD_GetFilePointer
                        DWORD   TD_SetFilePointer
                        DWORD   TD_ReadFile
                        DWORD   TD_WriteFile
                        DWORD   TD_RenameFile
                        DWORD   TD_DeleteFile
                        DWORD   TD_Hash                                 ; TRSDOS additional member functions
                        DWORD   TD_DEC2FDE
                        DWORD   TD_Sector2CHS

.data?

;---------------------------------------------------------------------
; TRSDOS Granule Allocation Table (GAT)
;---------------------------------------------------------------------

TD_GAT  struct
    AllocationTable     BYTE    96  dup (?)                             ; Free/Assigned table information
    LockoutTable        BYTE    96  dup (?)                             ; Used during mirror-image backups to determine if the destination disk has the proper capacity
    ExtendedTable       BYTE    11  dup (?)                             ; Used in hard drive configurations to extend the table from 00 to CA
    DosVersion          BYTE    ?                                       ; Operation System version used in formatting the disk
    CylinderExcess      BYTE    ?                                       ; Number of cylinders in excess of 35
    DiskConfig          BYTE    ?                                       ; Data specific to the formatting of the diskette
    PasswordHash        BYTE    2   dup (?)                             ; Master password hash
    DiskName            BYTE    8   dup (?)                             ; Disk name
    DiskDate            BYTE    8   dup (?)                             ; Disk date
    Reserved            BYTE    21  dup (?)                             ; Reserved for system use
    MDB                 BYTE    4   dup (?)                             ; Media Data Block (MDB) (0x03, 'LSI')
    DCT                 BYTE    7   dup (?)                             ; Last 7 bytes of the DCT when the media was formatted
TD_GAT  ends

;---------------------------------------------------------------------
; TRSDOS File Extent Element
;---------------------------------------------------------------------

TD_Extent   struct
    Cylinder            BYTE    ?                                       ; Cylinder value for the starting granule of the extent
    Granules            BYTE    ?                                       ; First granule (bits 5-7) and number of contiguous granules (bits 0-4)
TD_Extent   ends

;---------------------------------------------------------------------
; TRSDOS File Primary Directory Entry
;---------------------------------------------------------------------

TD_FPDE struct
    Attributes          BYTE        3 dup (?)                           ; File attributes
    EOF                 BYTE        ?                                   ; EOF offset byte
    LRL                 BYTE        ?                                   ; Logical Record Length (LRL)
    FileName            BYTE        8 dup (?)                           ; File name
    Extension           BYTE        3 dup (?)                           ; File extension
    OwnerHash           WORD        ?                                   ; OWNER password hash
    UserHash            WORD        ?                                   ; USER password hash
    ERN                 WORD        ?                                   ; Ending Record Number (ERN)
    Extent              TD_Extent   4 dup (<>)                          ; Four extent elements
    Link                TD_Extent   <>                                  ; Link to File Extended Directory Entry
TD_FPDE ends

;---------------------------------------------------------------------
; TRSDOS Class
;---------------------------------------------------------------------

TRSDOS  struct
    VPTR                DWORD   ?                                       ; Pointer to this object's VTBL
    pVDI                DWORD   ?                                       ; Pointer to an initialized VDI object
    pDir                DWORD   ?                                       ; Pointer to Directory data
    DirSectors          BYTE    ?                                       ; Number of directory sectors
    FirstTrack          BYTE    ?                                       ; Number of the first track on disk (0 or 1)
    FirstSide           BYTE    ?                                       ; Number of the first disk side (0 or 1)
    FirstSector         BYTE    ?                                       ; Number of the first sector on track (0 or 1)
    Sides               BYTE    ?                                       ; Number of disk sides (1 or 2)
    SectorsPerTrack     BYTE    ?                                       ; Number of sectors per track
    GranulesPerTrack    BYTE    ?                                       ; Number of granules per track
    SectorsPerGranule   BYTE    ?                                       ; Number of sectors per granule
    LastDEC             BYTE    ?                                       ; Last Directory Entry Code set by TD_FindFiles
    FilePointer         DWORD   ?                                       ; Current file pointer set by TD_SetFilePointer
    FileSector          WORD    ?                                       ; Current relative sector computed by TD_SetFilePointer
TRSDOS  ends

;---------------------------------------------------------------------
; TRSDOS Types
;---------------------------------------------------------------------

PTD                     typedef PTR TRSDOS

;---------------------------------------------------------------------
; TRSDOS Prototypes
;---------------------------------------------------------------------

TD_Load                 proto   :PTD, :PVDI
TD_Unload               proto   :PTD
TD_GetDosVersion        proto   :PTD
TD_GetDiskName          proto   :PTD, :DWORD, :DWORD
TD_SetDiskName          proto   :PTD, :DWORD
TD_GetDiskDate          proto   :PTD, :DWORD, :DWORD
TD_SetDiskDate          proto   :PTD, :DWORD
TD_FindFile             proto   :PTD, :BYTE
TD_GetFileHandle        proto   :PTD, :DWORD
TD_GetFileName          proto   :PTD, :DWORD, :DWORD, :DWORD
TD_GetFileExtension     proto   :PTD, :DWORD, :DWORD, :DWORD
TD_GetFileSize          proto   :PTD, :DWORD
TD_GetFileDate          proto   :PTD, :DWORD
TD_SetFileDate          proto   :PTD, :DWORD, :DWORD
TD_GetFileAttributes    proto   :PTD, :DWORD
TD_SetFileAttributes    proto   :PTD, :DWORD, :DWORD
TD_GetFilePointer       proto   :PTD, :DWORD
TD_SetFilePointer       proto   :PTD, :DWORD, :DWORD
TD_ReadFile             proto   :PTD, :DWORD, :DWORD, :DWORD
TD_WriteFile            proto   :PTD, :DWORD, :DWORD, :DWORD
TD_RenameFile           proto   :PTD, :DWORD, :DWORD
TD_DeleteFile           proto   :PTD, :DWORD
TD_Hash                 proto   :PTD, :DWORD
TD_DEC2FDE              proto   :PTD, :BYTE
TD_Sector2CHS           proto   :PTD, :WORD

.code

;---------------------------------------------------------------------
; TD_Load                                         Initialize interface
;---------------------------------------------------------------------
; pTD:      Pointer to the calling TRSDOS object
; pVDI:     Pointer to an initialized VDI object
;---------------------------------------------------------------------
; EAX:      Error code
;---------------------------------------------------------------------

TD_Load     proc    pTD:PTD, pVDI:PVDI

            local   Boot[256]:BYTE
            local   pDir:DWORD
            local   DirSectors:BYTE
            local   Side:BYTE
            local   Sector:BYTE
            local   Sides:BYTE
            local   SectorsPerTrack:BYTE
            local   FirstTrack:BYTE
            local   FirstSide:BYTE
            local   FirstSector:BYTE
            local   LastSector:BYTE
            local   GranulesPerTrack:BYTE
            local   SectorsPerGranule:BYTE

    ; --- Check to which function the object's VTBL points to ---

            mov     eax, pTD
            mov     eax, (TRSDOS PTR [eax]).VPTR

            cmp     [eax + TD_Call_Load], TD_Load
            je      @F

            mov     edx, pVDI
            push    edx
            mov     edx, pTD
            push    edx

            call    dword ptr [eax + TD_Call_Load]
            jmp     Done

    ; --- Get disk information ---

@@:         fn      VDI_GetSides, pVDI
            mov     Sides, al

            fn      VDI_GetSectorsPerTrack, pVDI
            mov     SectorsPerTrack, al

            fn      VDI_GetFirstTrack, pVDI
            mov     FirstTrack, al

            fn      VDI_GetFirstSide, pVDI
            mov     FirstSide, al

            fn      VDI_GetFirstSector, pVDI
            mov     FirstSector, al

            fn      VDI_GetLastSector, pVDI
            mov     LastSector, al

    ; --- Load the boot sector to find out the directory cylinder ---

            fn      VDI_ReadSector, pVDI, addr Boot, sizeof Boot, FirstTrack, FirstSide, FirstSector
            
            or      eax, eax
            jnz     Done

            cmp     Boot[1], 0FEh
            jne     IntError

    ; --- Compute the number of directory sectors ---

            mov     al, SectorsPerTrack                                 ; Directory Sectors = Sectors per Track * Sides
            mul     Sides

            cmp     al, 32+2                                            ; Max sectors limited by HIT: 256 hashes / 8 FDEs per sector = 32 sectors (+2 for GAT and HIT)
            jbe     @F

            mov     al, 32+2                                            ; Set top limit

@@:         mov     DirSectors, al
            
    ; --- Allocate memory for the entire directory ---

            shl     eax, 8                                              ; Buffer Size = DirSectors * 256 bytes per sector (TRSDOS standard)

            fn      GlobalAlloc, GPTR, eax

            or      eax, eax
            jnz     @F

            fn      GetLastError
            jmp     Done

@@:         mov     pDir, eax

    ; --- Load the directory into memory ---

            mov     edx, eax
            movzx   ecx, DirSectors

            mov     al, FirstSide
            mov     Side, al

            mov     al, FirstSector
            mov     Sector, al

Read:       push    ecx
            push    edx
            
            fn      VDI_ReadSector, pVDI, edx, 256, Boot[2], Side, Sector   ; Boot[2] contains the directory cylinder

            pop     edx
            pop     ecx
            
            or      eax, eax
            jnz     Done

            add     edx, 256                                            ; Advance pointer to the next memory block

            inc     Sector
            mov     al, Sector                                          ; Reached the end of the track?
            cmp     al, LastSector
            jbe     @F

            mov     al, FirstSector                                     ; Restart sector count
            mov     Sector, al

            inc     Side                                                ; Continue on the other side

@@:         loop    Read                                                ; Loop through all directory sectors

    ; --- Compute and validate the disk's granule units ---

            mov     edx, pDir                                           ; Pointer to the first directory sector (GAT)

            mov     cl, (TD_GAT PTR [edx]).DiskConfig                   ; Find the number of granules per track
            and     cl, TD_GAT_Granules
            inc     cl
            mov     GranulesPerTrack, cl

            cmp     cl, 2                                               ; Granules per Track should be between 2 and 8
            jb      IntError

            mov     al, SectorsPerTrack                                 ; Compute the number of sectors per granule
            div     cl
            mov     SectorsPerGranule, al

            or      ah, ah                                              ; There must be no remainder after last division
            jnz     IntError

            cmp     al, 5                                               ; Sectors per Granule for 5.25" SDEN disks
            je      @F
    
            cmp     al, 6                                               ; Sectors per Granule for 5.25" DDEN disks
            je      @F
    
            cmp     al, 8                                               ; Sectors per Granule for 8" SDEN disks
            je      @F
    
            cmp     al, 10                                              ; Sectors per Granule for 8" DDEN disks
            jne     IntError
    
    ; --- Update member class variables ---
    
@@:         mov     edx, pTD
            
            mov     eax, pVDI                                           ; Update pointer to Virtual Disk Interface
            mov     (TRSDOS PTR [edx]).pVDI, eax

            mov     eax, pDir                                           ; Update pointer to loaded Directory
            mov     (TRSDOS PTR [edx]).pDir, eax

            mov     al, DirSectors                                      ; Update number of directory sectors
            mov     (TRSDOS PTR [edx]).DirSectors, al

            mov     al, FirstTrack                                      ; Update number of the first track on disk
            mov     (TRSDOS PTR [edx]).FirstTrack, al

            mov     al, FirstSide                                       ; Update number of the first disk side
            mov     (TRSDOS PTR [edx]).FirstSide, al

            mov     al, FirstSector                                     ; Update number of the first track on disk
            mov     (TRSDOS PTR [edx]).FirstSector, al

            mov     al, Sides                                           ; Update number of disk sides
            mov     (TRSDOS PTR [edx]).Sides, al

            mov     al, SectorsPerTrack                                 ; Update number of Sectors per Track
            mov     (TRSDOS PTR [edx]).SectorsPerTrack, al

            mov     al, GranulesPerTrack                                ; Update number of Granules per Track
            mov     (TRSDOS PTR [edx]).GranulesPerTrack, al

            mov     al, SectorsPerGranule                               ; Update number of Sectors per Granule
            mov     (TRSDOS PTR [edx]).SectorsPerGranule, al

            mov     eax, NO_ERROR
            jmp     Done

IntError:   mov     eax, ERROR_NOT_DOS_DISK

Done:       ret

TD_Load     endp

;---------------------------------------------------------------------
; TD_Unload                                          Release interface
;---------------------------------------------------------------------
; pTD:      Pointer to the calling TRSDOS object
;---------------------------------------------------------------------
; EAX:      Error code
;---------------------------------------------------------------------

TD_Unload   proc    pTD:PTD

    ; --- Check to which function the object's VTBL points to ---

            mov     eax, pTD
            mov     eax, (TRSDOS PTR [eax]).VPTR

            cmp     [eax + TD_Call_Unload], TD_Unload
            je      @F

            mov     edx, pTD
            push    edx

            call    dword ptr [eax + TD_Call_Unload]
            jmp     Done

    ; --- Release allocated memory ---

@@:         mov     eax, pTD

            fn      GlobalFree, (TRSDOS PTR [eax]).pDir

            or      eax, eax
            jz      Done

            fn      GetLastError

Done:       ret

TD_Unload   endp

;---------------------------------------------------------------------
; TD_GetDosVersion                              Return the DOS version
;---------------------------------------------------------------------
; pTD:      Pointer to the calling TRSDOS object
;---------------------------------------------------------------------
; EAX:      DOS version
;---------------------------------------------------------------------

TD_GetDosVersion    proc    pTD:PTD

    ; --- Check to which function the object's VTBL points to ---

            mov     eax, pTD
            mov     eax, (TRSDOS PTR [eax]).VPTR

            cmp     [eax + TD_Call_GetDosVersion], TD_GetDosVersion
            je      @F

            mov     edx, pTD
            push    edx

            call    dword ptr [eax + TD_Call_GetDosVersion]
            jmp     Done

    ; --- Get DOS version from GAT sector ---
            
@@:         mov     eax, pTD
            mov     eax, (TRSDOS PTR [eax]).pDir
            movzx   eax, (TD_GAT PTR [eax]).DosVersion

Done:       ret
    
TD_GetDosVersion    endp

;---------------------------------------------------------------------
; TD_GetDiskName                                  Return the disk name
;---------------------------------------------------------------------
; pTD:      Pointer to the calling TRSDOS object
; pStr:     Pointer to the receiving string buffer
; Len:      Buffer length
;---------------------------------------------------------------------
; EAX:      Error Code
;---------------------------------------------------------------------

TD_GetDiskName      proc    pTD:PTD, pStr:DWORD, Len:DWORD

    ; --- Check to which function the object's VTBL points to ---

            mov     eax, pTD
            mov     eax, (TRSDOS PTR [eax]).VPTR

            cmp     [eax + TD_Call_GetDiskName], TD_GetDiskName
            je      @F

            mov     edx, Len
            push    edx
            mov     edx, pStr
            push    edx
            mov     edx, pTD
            push    edx

            call    dword ptr [eax + TD_Call_GetDiskName]
            jmp     Done
            
    ; --- Get disk name from GAT sector ---

@@:         cmp     Len, sizeof TD_GAT.DiskName
            jae     @F

            mov     eax, ERROR_INSUFFICIENT_BUFFER
            jmp     Done

@@:         mov     edx, pTD
            mov     edx, (TRSDOS PTR [edx]).pDir
            lea     edx, (TD_GAT PTR [edx]).DiskName

            fn      MemCopy, edx, pStr, sizeof TD_GAT.DiskName

            mov     eax, NO_ERROR

Done:       ret
    
TD_GetDiskName      endp

;---------------------------------------------------------------------
; TD_SetDiskName                                     Set the disk name
;---------------------------------------------------------------------
; pTD:      Pointer to the calling TRSDOS object
; pName:    Pointer to null-terminated string
;---------------------------------------------------------------------
; EAX:      Error code
;---------------------------------------------------------------------

TD_SetDiskName      proc    pTD:PTD, pName:DWORD

    ; --- Check to which function the object's VTBL points to ---

            mov     eax, pTD
            mov     eax, (TRSDOS PTR [eax]).VPTR

            cmp     [eax + TD_Call_SetDiskName], TD_SetDiskName
            je      @F

            mov     edx, pName
            push    edx
            mov     edx, pTD
            push    edx

            call    dword ptr [eax + TD_Call_SetDiskName]
            jmp     Done
            
    ; --- ---

@@:         mov     eax, NO_ERROR
    
Done:       ret
    
TD_SetDiskName      endp

;---------------------------------------------------------------------
; TD_GetDiskDate                                  Return the disk date
;---------------------------------------------------------------------
; pTD:      Pointer to the calling TRSDOS object
; pStr:     Pointer to the receiving string buffer
; Len:      Buffer length
;---------------------------------------------------------------------
; EAX:      Error code
;---------------------------------------------------------------------

TD_GetDiskDate      proc    pTD:PTD, pStr:DWORD, Len:DWORD

    ; --- Check to which function the object's VTBL points to ---

            mov     eax, pTD
            mov     eax, (TRSDOS PTR [eax]).VPTR

            cmp     [eax + TD_Call_GetDiskDate], TD_GetDiskDate
            je      @F

            mov     edx, Len
            push    edx
            mov     edx, pStr
            push    edx
            mov     edx, pTD
            push    edx

            call    dword ptr [eax + TD_Call_GetDiskDate]
            jmp     Done
            
    ; --- Get disk name from GAT sector ---

@@:         cmp     Len, sizeof TD_GAT.DiskDate
            jae     @F

            mov     eax, ERROR_INSUFFICIENT_BUFFER
            jmp     Done

@@:         mov     edx, pTD
            mov     edx, (TRSDOS PTR [edx]).pDir
            lea     edx, (TD_GAT PTR [edx]).DiskDate

            fn      MemCopy, edx, pStr, sizeof TD_GAT.DiskDate

            mov     eax, NO_ERROR

Done:       ret
    
TD_GetDiskDate      endp

;---------------------------------------------------------------------
; TD_SetDiskDate                                     Set the disk date
;---------------------------------------------------------------------
; pTD:      Pointer to the calling TRSDOS object
; pDate:    Pointer to null-terminated string
;---------------------------------------------------------------------
; EAX:      Error code
;---------------------------------------------------------------------

TD_SetDiskDate      proc    pTD:PTD, pDate:DWORD

    ; --- Check to which function the object's VTBL points to ---

            mov     eax, pTD
            mov     eax, (TRSDOS PTR [eax]).VPTR

            cmp     [eax + TD_Call_SetDiskDate], TD_SetDiskDate
            je      @F

            mov     edx, pDate
            push    edx
            mov     edx, pTD
            push    edx

            call    dword ptr [eax + TD_Call_SetDiskDate]
            jmp     Done
            
    ; --- ---

@@:         mov     eax, NO_ERROR
    
Done:       ret
    
TD_SetDiskDate      endp

;---------------------------------------------------------------------
; TD_FindFiles                           Search files in the directory
;---------------------------------------------------------------------
; pTD:      Pointer to the calling TRSDOS object
; Flag:     OSI_FINDFIRST or OSI_FINDNEXT
;---------------------------------------------------------------------
; EAX:      Handle to FPDE, NULL if no more files, -1 on error
;---------------------------------------------------------------------

TD_FindFiles        proc    pTD:PTD, Flag:BYTE

    ; --- Check to which function the object's VTBL points to ---

            mov     eax, pTD
            mov     eax, (TRSDOS PTR [eax]).VPTR

            cmp     [eax + TD_Call_FindFiles], TD_FindFiles
            je      @F

            movzx   edx, Flag
            push    edx
            mov     edx, pTD
            push    edx

            call    dword ptr [eax + TD_Call_FindFiles]
            jmp     Done
            
    ; --- Scan the Hash Index Table ---

@@:         mov     edx, pTD

            movzx   ecx, (TRSDOS PTR [edx]).LastDEC                     ; Get the last used Directory Entry Code
            inc     ecx

            cmp     Flag, OSI_FINDFIRST                                 ; Requested to look for first file?
            jne     @F

            mov     ecx, 0                                              ; Set the DEC back to zero

@@:         mov     edx, (TRSDOS PTR [edx]).pDir                        ; Point to the Hash Index Table
            add     edx, 256 * 1

Scan:       cmp     byte ptr [edx + ecx], 0                             ; Look for a non-empty HIT slot
            jne     Check

Next:       inc     ecx

            cmp     ecx, 256
            jb      Scan

            mov     eax, NULL                                           ; No more files found, return null
            jmp     Done

    ; --- Check the directory entry ---

Check:      push    ecx
            push    edx

            fn      TD_DEC2FDE, pTD, cl                                 ; Get pointer to directory entry based on DEC

            pop     edx
            pop     ecx

            cmp     eax, -1                                             ; Check if it's a valid pointer
            je      Done

            test    (TD_FPDE PTR [eax]).Attributes[0], TD_FPDE_Extended ; Check whether the entry is primary or extended
            jnz     Next

            test    (TD_FPDE PTR [eax]).Attributes[0], TD_FPDE_Active   ; Check if the FPDE is active
            jz      Next

            mov     edx, pTD
            mov     (TRSDOS PTR [edx]).LastDEC, cl                      ; Update the last DEC

Done:       ret
    
TD_FindFiles        endp

;---------------------------------------------------------------------
; TD_GetFileHandle                          Return handle by file name
;---------------------------------------------------------------------
; pTD:      Pointer to the calling TRSDOS object
; pFile:    Pointer to blank-padded file name and extension
;---------------------------------------------------------------------
; EAX:      Handle to FPDE, NULL if not found, -1 on error
;---------------------------------------------------------------------

TD_GetFileHandle    proc    pTD:PTD, pFile:DWORD

            local   hFPDE:DWORD

    ; --- Check to which function the object's VTBL points to ---

            mov     eax, pTD
            mov     eax, (TRSDOS PTR [eax]).VPTR

            cmp     [eax + TD_Call_GetFileHandle], TD_GetFileHandle
            je      @F

            mov     edx, pFile
            push    edx
            mov     edx, pTD
            push    edx

            call    dword ptr [eax + TD_Call_GetFileHandle]
            jmp     Done

    ; --- Look for the file hash code in the HIT ---

@@:         mov     hFPDE, 0

            fn      TD_Hash, pTD, pFile                                 ; Hash the file name

            mov     edx, pTD
            mov     edx, (TRSDOS PTR [edx]).pDir                        ; Point to the Hash Index Table
            add     edx, 256 * 1

            mov     ecx, 256                                            ; Scan length (256 slots)

Scan:       cmp     byte ptr [edx], al                                  ; Look for a matching hash code
            je      Check

Next:       inc     edx
            loop    Scan

            mov     eax, NULL                                           ; If not found, return null
            jmp     Done

    ; --- Check the directory entry ---

Check:      push    eax
            push    ecx
            push    edx
            push    esi
            push    edi

            neg     cx                                                  ; Directory Entry Code = ScanLength (256) - CX
            add     cx, 256

            fn      TD_DEC2FDE, pTD, cl                                 ; Get pointer to the File Primary Directory Entry

            cmp     eax, -1                                             ; Check if it's not an invalid pointer
            je      @F
            
            test    (TD_FPDE PTR [eax]).Attributes[0], TD_FPDE_Extended ; Check whether the entry is primary or extended
            jnz     Skip

            test    (TD_FPDE PTR [eax]).Attributes[0], TD_FPDE_Active   ; Check if the FPDE is active
            jz      Skip

            lea     esi, (TD_FPDE PTR [eax]).FileName                   ; Compare the two file names (check against collisions)
            mov     edi, pFile
            mov     ecx, 11
            cld
            repe    cmpsb
            jne     Skip

@@:         mov     hFPDE, eax
            
Skip:       pop     edi
            pop     esi
            pop     edx
            pop     ecx
            pop     eax

            cmp     hFPDE, 0                                            ; Keep searching while hFPDE is null
            je      Next

            mov     eax, hFPDE

Done:       ret
    
TD_GetFileHandle    endp

;---------------------------------------------------------------------
; TD_GetFileName                                  Return the file name
;---------------------------------------------------------------------
; pTD:      Pointer to the calling TRSDOS object
; hFPDE:    Pointer to the File Primary Directory Entry
; pStr:     Pointer to the receiving string buffer
; Len:      Buffer length
;---------------------------------------------------------------------
; EAX:      Error code
;---------------------------------------------------------------------

TD_GetFileName      proc    pTD:PTD, hFPDE:DWORD, pStr:DWORD, Len:DWORD

    ; --- Check to which function the object's VTBL points to ---

            mov     eax, pTD
            mov     eax, (TRSDOS PTR [eax]).VPTR

            cmp     [eax + TD_Call_GetFileName], TD_GetFileName
            je      @F

            mov     edx, Len
            push    edx
            mov     edx, pStr
            push    edx            
            mov     edx, hFPDE
            push    edx
            mov     edx, pTD
            push    edx

            call    dword ptr [eax + TD_Call_GetFileName]
            jmp     Done
            
    ; --- Copy file name from directory entry ---

@@:         cmp     Len, sizeof TD_FPDE.FileName
            jae     @F

            mov     eax, ERROR_INSUFFICIENT_BUFFER
            jmp     Done

@@:         mov     edx, hFPDE
            lea     edx, (TD_FPDE PTR [edx]).FileName

            fn      MemCopy, edx, pStr, sizeof TD_FPDE.FileName

            mov     eax, NO_ERROR

Done:       ret
    
TD_GetFileName      endp

;---------------------------------------------------------------------
; TD_GetFileExtension                        Return the file extension
;---------------------------------------------------------------------
; pTD:      Pointer to the calling TRSDOS object
; hFPDE:    Pointer to the File Primary Directory Entry
; pStr:     Pointer to the receiving string buffer
; Len:      Buffer length
;---------------------------------------------------------------------
; EAX:      Error code
;---------------------------------------------------------------------

TD_GetFileExtension proc    pTD:PTD, hFPDE:DWORD, pStr:DWORD, Len:DWORD

    ; --- Check to which function the object's VTBL points to ---

            mov     eax, pTD
            mov     eax, (TRSDOS PTR [eax]).VPTR

            cmp     [eax + TD_Call_GetFileExtension], TD_GetFileExtension
            je      @F

            mov     edx, Len
            push    edx
            mov     edx, pStr
            push    edx            
            mov     edx, hFPDE
            push    edx
            mov     edx, pTD
            push    edx

            call    dword ptr [eax + TD_Call_GetFileExtension]
            jmp     Done
            
    ; --- Copy file extension from directory entry ---

@@:         cmp     Len, sizeof TD_FPDE.Extension
            jae     @F

            mov     eax, ERROR_INSUFFICIENT_BUFFER
            jmp     Done

@@:         mov     edx, hFPDE
            lea     edx, (TD_FPDE PTR [edx]).Extension

            fn      MemCopy, edx, pStr, sizeof TD_FPDE.Extension

            mov     eax, NO_ERROR

Done:       ret
    
TD_GetFileExtension endp

;---------------------------------------------------------------------
; TD_GetFileSize                                  Return the file size
;---------------------------------------------------------------------
; pTD:      Pointer to the calling TRSDOS object
; hFPDE:    Pointer to the File Primary Directory Entry
;---------------------------------------------------------------------
; EAX:      File size
;---------------------------------------------------------------------

TD_GetFileSize      proc    pTD:PTD, hFPDE:DWORD

    ; --- Check to which function the object's VTBL points to ---

            mov     eax, pTD
            mov     eax, (TRSDOS PTR [eax]).VPTR

            cmp     [eax + TD_Call_GetFileSize], TD_GetFileSize
            je      @F

            mov     edx, hFPDE
            push    edx
            mov     edx, pTD
            push    edx

            call    dword ptr [eax + TD_Call_GetFileSize]
            jmp     Done
            
    ; --- Compute the file size ---

@@:         mov     edx, hFPDE

            movzx   eax, (TD_FPDE PTR [edx]).ERN                        ; Get number of sectors (Ending Record Number) and multiply by 256 bytes
            shl     eax, 8
            
            movzx   ecx, (TD_FPDE PTR [edx]).EOF                        ; Get End of File (EOF) value
            jcxz    Done

            or      eax, eax                                            ; ERN = 0? 
            jnz     @F

            mov     eax, ecx                                            ; Size = EOF
            jmp     Done

@@:         neg     cl                                                  ; Subtract (256 - EOF) from the total
            sub     eax, ecx
            
Done:       ret
    
TD_GetFileSize      endp

;---------------------------------------------------------------------
; TD_GetFileDate                                  Return the file date
;---------------------------------------------------------------------
; pTD:      Pointer to the calling TRSDOS object
; hFPDE:    Pointer to the File Primary Directory Entry
;---------------------------------------------------------------------
; EAX:      Date (YYYYMMDD)
;---------------------------------------------------------------------

TD_GetFileDate      proc    pTD:PTD, hFPDE:DWORD

    ; --- Check to which function the object's VTBL points to ---

            mov     eax, pTD
            mov     eax, (TRSDOS PTR [eax]).VPTR

            cmp     [eax + TD_Call_GetFileDate], TD_GetFileDate
            je      @F

            mov     edx, hFPDE
            push    edx
            mov     edx, pTD
            push    edx

            call    dword ptr [eax + TD_Call_GetFileDate]
            jmp     Done
            
    ; --- Format the return value with the file date ---

@@:         mov     edx, hFPDE
            
            mov     cl, (TD_FPDE PTR [edx]).Attributes[2]               ; Get second attribute byte which contains the Year and Day
            
            mov     al, cl                                              ; Extract Year
            and     eax, TD_FPDE_Year
            add     eax, 1980
            shl     eax, 16                                             ; Put Year in EAX's MSB

            and     cl, TD_FPDE_Day                                     ; Extract Day    
            shr     cl, 3
            jnz     @F
            
            inc     cl                                                  ; If day is zero, set to one

@@:         mov     al, cl                                              ; Put Day in AL

            mov     cl, (TD_FPDE PTR [edx]).Attributes[1]               ; Get first attribute byte and extract the Month
            and     cl, TD_FPDE_Month
            jnz     @F
            
            inc     cl                                                  ; If month is zero, set to one
            
@@:         mov     ah, cl                                              ; Put Month in AH

Done:       ret
    
TD_GetFileDate      endp

;---------------------------------------------------------------------
; TD_SetFileDate                                  Change the file date
;---------------------------------------------------------------------
; pTD:      Pointer to the calling TRSDOS object
; hFPDE:    Pointer to the File Primary Directory Entry
; Date:     New file date (YYYYMMDD)
;---------------------------------------------------------------------
; EAX:      Error code
;---------------------------------------------------------------------

TD_SetFileDate      proc    pTD:PTD, hFPDE:DWORD, Date:DWORD

    ; --- Check to which function the object's VTBL points to ---

            mov     eax, pTD
            mov     eax, (TRSDOS PTR [eax]).VPTR

            cmp     [eax + TD_Call_SetFileDate], TD_SetFileDate
            je      @F

            mov     edx, Date
            push    edx
            mov     edx, hFPDE
            push    edx
            mov     edx, pTD
            push    edx

            call    dword ptr [eax + TD_Call_SetFileDate]
            jmp     Done
            
    ; --- ---

@@:         mov     eax, NO_ERROR
    
Done:       ret
    
TD_SetFileDate      endp

;---------------------------------------------------------------------
; TD_GetFileAttributes                      Return the file attributes
;---------------------------------------------------------------------
; pTD:      Pointer to the calling TRSDOS object
; hFPDE:    Pointer to the File Primary Directory Entry
;---------------------------------------------------------------------
; EAX:      File attributes (0000MISA)
;---------------------------------------------------------------------

TD_GetFileAttributes    proc    pTD:PTD, hFPDE:DWORD

    ; --- Check to which function the object's VTBL points to ---

            mov     eax, pTD
            mov     eax, (TRSDOS PTR [eax]).VPTR

            cmp     [eax + TD_Call_GetFileAttributes], TD_GetFileAttributes
            je      @F

            mov     edx, hFPDE
            push    edx
            mov     edx, pTD
            push    edx

            call    dword ptr [eax + TD_Call_GetFileAttributes]
            jmp     Done
            
    ; --- Format the return value with the file attributes ---

@@:         mov     edx, hFPDE

            movzx   eax, (TD_FPDE PTR [edx]).Attributes[0]              ; Get the file attributes
            
            and     ax, TD_FPDE_Access                                  ; Strip the file access level

            test    (TD_FPDE PTR [edx]).Attributes[0], TD_FPDE_System       ; Test the System attribute
            jz      @F

            bts     eax, 4
            
@@:         test    (TD_FPDE PTR [edx]).Attributes[0], TD_FPDE_Invisible    ; Test the Invisible attribute
            jz      @F

            bts     eax, 8
            
@@:         test    (TD_FPDE PTR [edx]).Attributes[1], TD_FPDE_Modified     ; Test the Modified attribute
            jz      Done

            bts     eax, 12
            
Done:       ret
    
TD_GetFileAttributes    endp

;---------------------------------------------------------------------
; TD_SetFileAttributes                      Change the file attributes
;---------------------------------------------------------------------
; pTD:      Pointer to the calling TRSDOS object
; hFPDE:    Pointer to the File Primary Directory Entry
; Attr:     New file attributes (0000MISA)
;---------------------------------------------------------------------
; EAX:      Error code
;---------------------------------------------------------------------

TD_SetFileAttributes    proc    pTD:PTD, hFPDE:DWORD, Attr:DWORD

    ; --- Check to which function the object's VTBL points to ---

            mov     eax, pTD
            mov     eax, (TRSDOS PTR [eax]).VPTR

            cmp     [eax + TD_Call_SetFileAttributes], TD_SetFileAttributes
            je      @F

            mov     edx, Attr
            push    edx
            mov     edx, hFPDE
            push    edx
            mov     edx, pTD
            push    edx

            call    dword ptr [eax + TD_Call_SetFileAttributes]
            jmp     Done
            
    ; --- ---

@@:         mov     eax, NO_ERROR
    
Done:       ret
    
TD_SetFileAttributes    endp

;---------------------------------------------------------------------
; TD_GetFilePointer                        Return current file pointer
;---------------------------------------------------------------------
; pTD:      Pointer to the calling TRSDOS object
; hFPDE:    Handle to File Primary Directory Entry
;---------------------------------------------------------------------
; EAX:      Current pointer position
;---------------------------------------------------------------------

TD_GetFilePointer   proc    pTD:PTD, hFPDE:DWORD

    ; --- Check to which function the object's VTBL points to ---

            mov     eax, pTD
            mov     eax, (TRSDOS PTR [eax]).VPTR

            cmp     [eax + TD_Call_GetFilePointer], TD_GetFilePointer
            je      @F

            mov     edx, hFPDE
            push    edx
            mov     edx, pTD
            push    edx

            call    dword ptr [eax + TD_Call_GetFilePointer]
            jmp     Done
            
    ; --- Return the stored file pointer value ---

@@:         mov     eax, pTD
            mov     eax, (TRSDOS PTR [eax]).FilePointer

Done:       ret
    
TD_GetFilePointer   endp

;---------------------------------------------------------------------
; TD_SetFilePointer                                  Move file pointer
;---------------------------------------------------------------------
; pTD:      Pointer to the calling TRSDOS object
; hFPDE:    Handle to File Primary Directory Entry
; Pos:      New pointer position
;---------------------------------------------------------------------
; EAX:      Error code
;---------------------------------------------------------------------

TD_SetFilePointer   proc    pTD:PTD, hFPDE:DWORD, Pos:DWORD

            local   SectorsPerGranule:BYTE
            local   Sector:WORD

    ; --- Check to which function the object's VTBL points to ---

            mov     eax, pTD
            mov     eax, (TRSDOS PTR [eax]).VPTR

            cmp     [eax + TD_Call_SetFilePointer], TD_SetFilePointer
            je      @F

            mov     edx, Pos
            push    edx
            mov     edx, hFPDE
            push    edx
            mov     edx, pTD
            push    edx

            call    dword ptr [eax + TD_Call_SetFilePointer]
            jmp     Done
            
    ; --- Initialize variables ---

@@:         mov     eax, pTD
            mov     al, (TRSDOS PTR [eax]).SectorsPerGranule
            mov     SectorsPerGranule, al

    ; --- Compute the sector offset ---

            mov     eax, Pos                                            ; Sector = Pos / 256 bytes per sector
            shr     eax, 8
            mov     Sector, ax

    ; --- Find the extent that contains such sector ---

            mov     eax, hFPDE

Restart:    lea     edx, (TD_FPDE PTR [eax]).Extent

            mov     ecx, 5

Get:        mov     ax, word ptr [edx]                                  ; Get extent element

            cmp     al, 255                                             ; 255 means no more extent elements for this file
            je      EOF

            cmp     al, 254                                             ; 254 means other byte contains a DEC linking to a FXDE
            je      Link

            and     ah, 00011111y                                       ; Get the number of contiguous granules
            inc     ah
            shr     ax, 8
            mul     SectorsPerGranule                                   ; Convert granules into sectors

            cmp     ax, Sector                                          ; Check if the requested sector is in the current extent
            ja      Found

            sub     Sector, ax

Next:       add     edx, 2                                              ; Advance to the next extent
            loop    Get

EOF:        mov     eax, ERROR_HANDLE_EOF
            jmp     Done

Link:       fn      TD_DEC2FDE, pTD, ah                                 ; Get pointer to the FXDE

            cmp     eax, -1                                             ; Check if it's a valid pointer
            jne     Restart

            mov     eax, ERROR_FILE_CORRUPT
            jmp     Done

    ; --- Compute the relative sector ---

Found:      mov     cx, word ptr [edx]                                  ; Get extent element

            mov     edx, pTD

            movzx   eax, (TRSDOS PTR [edx]).GranulesPerTrack            ; Sector = (Cylinder * (GranulesPerTrack * Sides) + InitialGranule) * SectorsPerGranule
            mul     (TRSDOS PTR [edx]).Sides
            mul     cl                                                  ; CL = Cylinder

            and     ch, 11100000y                                       ; Add the extent's initial granule
            shr     cx, 5+8
            add     ax, cx

            movzx   cx, (TRSDOS PTR [edx]).SectorsPerGranule            ; Convert granules into sectors
            mul     cx

            add     Sector, ax                                          ; Add to the remaining sector offset

    ; --- Update member variables ---

            mov     edx, pTD

            mov     eax, Pos
            mov     (TRSDOS PTR [edx]).FilePointer, eax

            mov     ax, Sector
            mov     (TRSDOS PTR [edx]).FileSector, ax

            mov     eax, NO_ERROR

Done:       ret
            
TD_SetFilePointer   endp

;---------------------------------------------------------------------
; TD_ReadFile                                      Read data from file
;---------------------------------------------------------------------
; pTD:      Pointer to the calling TRSDOS object
; hFPDE:    Handle to File Primary Directory Entry
; pBuffer:  Pointer to the receiving buffer
; Bytes:    Number of bytes to read
;---------------------------------------------------------------------
; EAX:      Error code
;---------------------------------------------------------------------

TD_ReadFile         proc    pTD:PTD, hFPDE:DWORD, pBuffer:DWORD, Bytes:DWORD

            local   pVDI:DWORD
            local   Buffer[256]:BYTE
            local   LastPointer:DWORD

    ; --- Check to which function the object's VTBL points to ---

            mov     eax, pTD
            mov     eax, (TRSDOS PTR [eax]).VPTR

            cmp     [eax + TD_Call_ReadFile], TD_ReadFile
            je      Read

            mov     edx, Bytes
            push    edx
            mov     edx, pBuffer
            push    edx
            mov     edx, hFPDE
            push    edx
            mov     edx, pTD
            push    edx

            call    dword ptr [eax + TD_Call_ReadFile]
            jmp     Done
            
    ; --- Read the sector pointed by FileSector ---

Read:       mov     edx, pTD

            mov     eax, (TRSDOS PTR [edx]).pVDI
            mov     pVDI, eax

            mov     eax, (TRSDOS PTR [edx]).FilePointer
            mov     LastPointer, eax

            fn      TD_Sector2CHS, pTD, (TRSDOS PTR [edx]).FileSector

            fn      VDI_ReadSector, pVDI, addr Buffer, sizeof Buffer, cl, dh, dl

            or      eax, eax
            jnz     Done

    ; --- Copy data from internal buffer ---

            mov     ecx, LastPointer
            and     ecx, 000000FFh

            lea     edx, Buffer
            add     edx, ecx

            neg     ecx
            add     ecx, 256

            cmp     ecx, Bytes
            jbe     @F

            mov     ecx, Bytes

@@:         push    ecx
            fn      MemCopy, edx, pBuffer, ecx
            pop     ecx

    ; --- Advance to the next block and loop until all bytes are read ---

            add     pBuffer, ecx
            sub     Bytes, ecx
            jnz     Next

            mov     eax, NO_ERROR
            jmp     Done

Next:       add     ecx, LastPointer

            fn      TD_SetFilePointer, pTD, hFPDE, ecx

            or      eax, eax
            jz      Read

Done:       ret
    
TD_ReadFile         endp

;---------------------------------------------------------------------
; TD_WriteFile                                      Write data to file
;---------------------------------------------------------------------
; pTD:      Pointer to the calling TRSDOS object
; hFPDE:    Handle to File Primary Directory Entry
; pBuffer:  Pointer to the data buffer
; Bytes:    Number of bytes to write
;---------------------------------------------------------------------
; EAX:      Error code
;---------------------------------------------------------------------

TD_WriteFile        proc    pTD:PTD, hFPDE:DWORD, pBuffer:DWORD, Bytes:DWORD

    ; --- Check to which function the object's VTBL points to ---

            mov     eax, pTD
            mov     eax, (TRSDOS PTR [eax]).VPTR

            cmp     [eax + TD_Call_WriteFile], TD_WriteFile
            je      @F

            mov     edx, Bytes
            push    edx
            mov     edx, pBuffer
            push    edx
            mov     edx, hFPDE
            push    edx
            mov     edx, pTD
            push    edx

            call    dword ptr [eax + TD_Call_WriteFile]
            jmp     Done
            
    ; --- ---

@@:         mov     eax, NO_ERROR
    
Done:       ret
    
TD_WriteFile        endp

;---------------------------------------------------------------------
; TD_RenameFile                                            Rename file
;---------------------------------------------------------------------
; pTD:      Pointer to the calling TRSDOS object
; pOldName: Current file name
; pNewName: Desired file name 
;---------------------------------------------------------------------
; EAX:      Error code
;---------------------------------------------------------------------

TD_RenameFile       proc    pTD:PTD, pOldName:DWORD, pNewName:DWORD

    ; --- Check to which function the object's VTBL points to ---

            mov     eax, pTD
            mov     eax, (TRSDOS PTR [eax]).VPTR

            cmp     [eax + TD_Call_RenameFile], TD_RenameFile
            je      @F

            mov     edx, pNewName
            push    edx
            mov     edx, pOldName
            push    edx
            mov     edx, pTD
            push    edx

            call    dword ptr [eax + TD_Call_RenameFile]
            jmp     Done
            
    ; --- ---

@@:         mov     eax, NO_ERROR
    
Done:       ret
    
TD_RenameFile       endp

;---------------------------------------------------------------------
; TD_DeleteFile                                            Delete file
;---------------------------------------------------------------------
; pTD:      Pointer to the calling TRSDOS object
; pFile:    Name of the file to delete
;---------------------------------------------------------------------
; EAX:      Error code
;---------------------------------------------------------------------

TD_DeleteFile       proc    pTD:PTD, pFile:DWORD

    ; --- Check to which function the object's VTBL points to ---

            mov     eax, pTD
            mov     eax, (TRSDOS PTR [eax]).VPTR

            cmp     [eax + TD_Call_DeleteFile], TD_DeleteFile
            je      @F

            mov     edx, pFile
            push    edx
            mov     edx, pTD
            push    edx

            call    dword ptr [eax + TD_Call_DeleteFile]
            jmp     Done
            
    ; --- ---

@@:         mov     eax, NO_ERROR
    
Done:       ret
    
TD_DeleteFile       endp

;---------------------------------------------------------------------
; TD_Hash                                             Hash a file name
;---------------------------------------------------------------------
; pTD:      Pointer to the calling TRSDOS object
; pFile:    Pointer to a blank-padded file name and extension
;---------------------------------------------------------------------
; AL:       Computed hash code
;---------------------------------------------------------------------

TD_Hash     proc    pTD:PTD, pFile:DWORD

    ; --- Check to which function the object's VTBL points to ---

            mov     eax, pTD
            mov     eax, (TRSDOS PTR [eax]).VPTR

            cmp     [eax + TD_Call_Hash], TD_Hash
            je      @F

            mov     edx, pFile
            push    edx
            mov     edx, pTD
            push    edx

            call    dword ptr [eax + TD_Call_Hash]
            jmp     Done
            
    ; --- Compute the hash code for the file name ---

@@:         mov     edx, pFile
            mov     ecx, 11
            xor     eax, eax
            
@@:         xor     al, byte ptr [edx]
            inc     edx
            rol     al, 1
            loop    @B
            
            or      al, al
            jnz     Done
            
            inc     al

Done:       ret

TD_Hash     endp

;---------------------------------------------------------------------
; TD_DEC2FDE                    Return pointer to File Directory Entry
;---------------------------------------------------------------------
; pTD:      Pointer to the calling TRSDOS object
; DEC:      Directory Entry Code (DEC)
;---------------------------------------------------------------------
; EAX:      Pointer to directory entry or -1 on error
;---------------------------------------------------------------------

TD_DEC2FDE  proc    pTD:PTD, DEC$:BYTE

    ; --- Check to which function the object's VTBL points to ---

            mov     eax, pTD
            mov     eax, (TRSDOS PTR [eax]).VPTR

            cmp     [eax + TD_Call_DEC2FDE], TD_DEC2FDE
            je      @F

            movzx   edx, DEC$
            push    edx
            mov     edx, pTD
            push    edx

            call    dword ptr [eax + TD_Call_DEC2FDE]
            jmp     Done
            
    ; --- Compute pointer to the directory entry ---

@@:         mov     edx, pTD
            movzx   eax, DEC$
            
            mov     ah, al
            and     ah, 00011111y                                       ; Strip column bits to get the sector offset
            and     al, 11100000y                                       ; Strip row bits to get the FDE offset (already multiplied by 32 bytes)

            cmp     ah, (TRSDOS PTR [edx]).DirSectors                   ; Check whether this is will be a valid pointer
            jb      @F

            mov     eax, -1                                             ; Return -1 if not
            jmp     Done

@@:         add     eax, (TRSDOS PTR [edx]).pDir                        ; Make a pointer to the directory
            add     eax, 256 * 2                                        ; Skip GAT and HIT sectors

Done:       ret

TD_DEC2FDE  endp

;---------------------------------------------------------------------
; TD_Sector2CHS                       Convert relative sector to C/H/S
;---------------------------------------------------------------------
; pTD:      Pointer to the calling TRSDOS object
; Sector:   Relative sector number
;---------------------------------------------------------------------
; CL:       Cylinder
; DH:       Head
; DL:       Sector
;---------------------------------------------------------------------

TD_Sector2CHS       proc    pTD:PTD, Sector:WORD

    ; --- Check to which function the object's VTBL points to ---

            mov     eax, pTD
            mov     eax, (TRSDOS PTR [eax]).VPTR

            cmp     [eax + TD_Call_Sector2CHS], TD_Sector2CHS
            je      @F

            movzx   edx, Sector
            push    edx
            mov     edx, pTD
            push    edx

            call    dword ptr [eax + TD_Call_Sector2CHS]
            jmp     Done
            
    ; --- Compute pointer to the directory entry ---

@@:         mov     edx, pTD

            mov     al, (TRSDOS PTR [edx]).SectorsPerTrack              ; SectorsPerTrack * Sides
            mul     (TRSDOS PTR [edx]).Sides

            mov     cx, ax
            mov     ax, Sector                                          ; Cylinder (AL) = (RelSector / (SectorsPerTrack * Sides)) + FirstTrack
            div     cl                                                  ; AH = Remainder
            add     al, (TRSDOS PTR [edx]).FirstTrack

            mov     cl, al                                              ; CL = Cylinder
            
            movzx   ax, ah                                              ; Head (AL) = (Remainder / SectorsPerTrack) + FirstSide
            div     (TRSDOS PTR [edx]).SectorsPerTrack                  ; Sector (AH) = Remainder + FirstSector
            add     al, (TRSDOS PTR [edx]).FirstSide
            add     ah, (TRSDOS PTR [edx]).FirstSector

            mov     dh, al                                              ; DH = Head
            mov     dl, ah                                              ; DL = Sector

Done:       ret

TD_Sector2CHS       endp

