;-----------------------------------------------------------
; VDI Module Checking Tool        Copyright (C) Miguel Dutra
;-----------------------------------------------------------

.386
.model flat, stdcall
option casemap :none

include \masm32\include\windows.inc
include \masm32\include\kernel32.inc
include \masm32\include\user32.inc
include \masm32\include\masm32.inc
include \masm32\macros\macros.asm

includelib \masm32\lib\kernel32.lib
includelib \masm32\lib\user32.lib
includelib \masm32\lib\masm32.lib

include ..\lib\macros.asm
include ..\lib\util.asm
include ..\VDI\VDI.asm                                                  ; Patch
include ..\DMK\DMK.asm                                                  ; Patch

DiskDump        proto   :DWORD

.code

;-----------------------------------------------------------
; Entry Point
;-----------------------------------------------------------

Main        proc

            local   argv[16]:DWORD
            local   argc:BYTE
            local   pVDI:DWORD
            local   hFile:HANDLE
            local   Buffer[256]:BYTE
            local   Tracks:BYTE
            local   Sides:BYTE
            local   SPT:BYTE
            local   Density:BYTE
            local   FirstTrack:BYTE
            local   LastTrack:BYTE
            local   FirstSide:BYTE
            local   LastSide:BYTE
            local   FirstSector:BYTE
            local   LastSector:BYTE

    ; --- Open the image file --- 
    
            fnc     StdOut, "Opening file...\n"

            fn      GetCommandLine
            
            fn      MakeArgList, addr argv, addr argc, eax
            
            fn      CreateFile, argv[1*DWORD], GENERIC_READ+GENERIC_WRITE, 0, NULL, OPEN_EXISTING, 0, NULL

            cmp     eax, INVALID_HANDLE_VALUE
            jne     @F

            fn      GetLastError
            fn      PrintError, eax
            jmp     Done

@@:         mov     hFile, eax

    ; --- Create and initialize the VDI object --- 
    
            fnc     StdOut, "Creating VDI object...\n"

            fn      GlobalAlloc, GPTR, 256

            or      eax, eax
            jnz     @F
            
            fn      GetLastError
            fn      PrintError, eax
            jmp     Done
                     
@@:         mov     pVDI, eax

            mov     (CVDI PTR [eax]).VPTR, offset DMK                   ; Patch

    ; --- Initialize the interface --- 

            fnc     StdOut, "Loading interface...\n"
            
            mcall   VDI_VTBL, Load, pVDI, hFile, 0

            or      eax, eax
            jz      @F
            
            fn      PrintError, eax
            jmp     Destroy

    ; --- Retrieve the disk geometry --- 

@@:         fnc     StdOut, "Retrieving disk geometry...\n"
            
            mcall   VDI_VTBL, GetTracks, pVDI
            mov     Tracks, al

            mcall   VDI_VTBL, GetSides, pVDI
            mov     Sides, al

            mcall   VDI_VTBL, GetSPT, pVDI
            mov     SPT, al

            mcall   VDI_VTBL, GetDensity, pVDI
            mov     Density, al

            mcall   VDI_VTBL, GetFirstSector, pVDI
            mov     FirstSector, al

            mcall   VDI_VTBL, GetLastSector, pVDI
            mov     LastSector, al

            mcall   VDI_VTBL, GetFirstSide, pVDI
            mov     FirstSide, al

            mcall   VDI_VTBL, GetLastSide, pVDI
            mov     LastSide, al

            mcall   VDI_VTBL, GetFirstTrack, pVDI
            mov     FirstTrack, al

            mcall   VDI_VTBL, GetLastTrack, pVDI
            mov     LastTrack, al

            movzx   eax, Density
            push    eax
            movzx   eax, LastSector
            push    eax
            movzx   eax, FirstSector
            push    eax
            movzx   eax, SPT
            push    eax
            movzx   eax, LastSide
            push    eax
            movzx   eax, FirstSide
            push    eax
            movzx   eax, Sides
            push    eax
            movzx   eax, LastTrack
            push    eax
            movzx   eax, FirstTrack
            push    eax
            movzx   eax, Tracks
            push    eax

            fnc     wsprintf, addr Buffer, "Tracks=%d (%d:%d) Sides=%d (%d:%d) Sectors=%d (%d:%d) Density=%d\n"
            
            fn      StdOut, addr Buffer

            add     esp, 4*10
            
    ; --- DiskDump all disk sectors --- 

            fnc     StdOut, "Dumping disk sectors...\n"

            fn      DiskDump, pVDI

    ; --- Release the interface --- 

            fnc     StdOut, "Unloading interface...\n"

            mcall   VDI_VTBL, Unload, pVDI

            or      eax, eax
            jz      Destroy
            
            fn      PrintError, eax

    ; --- Delete the VDI object --- 

Destroy:    fnc     StdOut, "Destroying VDI object...\n"
            
            fn      GlobalFree, pVDI

Done:       fn      ExitProcess, 0

Main        endp

;-----------------------------------------------------------
; DiskDump disk sectors
;-----------------------------------------------------------

DiskDump    proc    pVDI:DWORD

            local   Buffer[1024]:BYTE
            local   String[64]:BYTE
            local   LastTrack:BYTE
            local   FirstSector:BYTE
            local   LastSector:BYTE
            local   FirstSide:BYTE
            local   LastSide:BYTE
            local   Track:BYTE
            local   Sector:BYTE
            local   Side:BYTE

    ; --- Retrieve the disk geometry ---

            mcall   VDI_VTBL, GetFirstTrack, pVDI
            mov     Track, al

            mcall   VDI_VTBL, GetLastTrack, pVDI
            mov     LastTrack, al

            mcall   VDI_VTBL, GetFirstSector, pVDI
            mov     FirstSector, al
            mov     Sector, 0                                           ; Always start from 0 (some disks have first track formatted differently)

            mcall   VDI_VTBL, GetLastSector, pVDI
            mov     LastSector, al

            mcall   VDI_VTBL, GetFirstSide, pVDI
            mov     FirstSide, al
            mov     Side, al

            mcall   VDI_VTBL, GetLastSide, pVDI
            mov     LastSide, al

    ; --- Read and dump the sector ---

DiskDumpSector: mcall   VDI_VTBL, Read, pVDI, addr Buffer, sizeof Buffer, Track, Side, Sector    ; Read sector at first track, first side, sector 0

            or      eax, eax                                            ; On error, skip the sector
            jnz     Next

            push    ecx

            movzx   edx, Track
            movzx   ecx, Side
            movzx   eax, Sector
            
            fn      wsprintf, addr String, "%c%c[%02u:%u:%02u]%c%c", 0Dh, 0Ah, edx, ecx, eax, 0Dh, 0Ah
    
            fn      StdOut, addr String                                 ; Print the sector address (CC:H:SS)

            pop     ecx

            fn      Dump, addr Buffer, ecx                              ; Print sector data

    ; --- Advance to next Sector, Side, Track ---

Next:       inc     Sector                                              ; Increment sector
            mov     al, Sector
            cmp     al, LastSector
            jbe     DiskDumpSector
        
            mov     al, FirstSector                                     ; Reset sector count
            mov     Sector, al

            inc     Side                                                ; Increment side
            mov     al, Side
            cmp     al, LastSide
            jbe     DiskDumpSector

            mov     al, FirstSide                                       ; Reset side count
            mov     Side, al

            inc     Track                                               ; Increment track
            mov     al, Track
            cmp     al, LastTrack
            jbe     DiskDumpSector

Done:       ret
    
DiskDump    endp

end Main

