;---------------------------------------------------------------------
; Graphical User Interface             Copyright (C) 2009 Miguel Dutra
;---------------------------------------------------------------------

GUI_Export  proto   :HWND

.data

ExportTitle db      "Select a directory to receive the exported files:", 0

.code

;---------------------------------------------------------------------
; GUI_Export                                   Export dialog procedure
;---------------------------------------------------------------------

GUI_Export  proc    hWnd:HWND

            local   hControl:HWND
            local   BI:BROWSEINFO
            local   LVI:LV_ITEM
            local   Path[MAX_PATH]:BYTE
            local   FileName[9]:BYTE
            local   FileExt[4]:BYTE
            local   FileSize:DWORD
            local   pBuffer:DWORD
            local   hFPDE:DWORD
            local   Index:DWORD
            local   TrsName[12]:BYTE
            local   WinName[13]:BYTE
            local   File[MAX_PATH]:BYTE
            local   hFile:HANDLE
            local   Bytes:DWORD

    ; --- Check whether there are any selected file ---

            fn      GetDlgItem, hWnd, IDC_LIST
            mov     hControl, eax

            fn      SendMessage, hControl, LVM_GETSELECTEDCOUNT, 0, 0

            or      eax, eax
            jnz     @F

            fn      MessageBox, hWnd, "No files selected.", "Export", MB_ICONINFORMATION+MB_OK
            jmp     Done

    ; --- Open the Browse Folder dialog ---

@@:         mov     eax, hWnd
            mov     BI.hwndOwner, eax
            mov     BI.pidlRoot, NULL
            lea     eax, Path
            mov     BI.pszDisplayName, eax
            mov     BI.lpszTitle, offset ExportTitle
            mov     BI.ulFlags, 0
            mov     BI.lpfn, NULL
            mov     BI.lParam, 0
            lea     eax, Bytes
            mov     BI.iImage, eax
            
            fn      SHBrowseForFolder, addr BI

            or      eax, eax
            jz      Done

    ; --- Get path from ID List ---

            mov     edx, eax

            fn      SHGetPathFromIDList, edx, addr Path

            or      eax, eax
            jnz     @F

            fn      MessageBox, hWnd, "SHGetPathFromIDList()", "Error", MB_ICONERROR+MB_OK
            jmp     Done

@@:         fn      szCatStr, addr Path, "\"                            ; Append a backslash to the end of the path

    ; --- Allocate memory for file contents ---

            fn      GlobalAlloc, GPTR, 80*30*2*256                      ; 1.2MB should be enough for most TRS-80 files

            or      eax, eax
            jnz     @F

            fn      MessageBox, hWnd, "GlobalAlloc()", "Error", MB_ICONERROR+MB_OK
            jmp     Done

@@:         mov     pBuffer, eax

    ; --- Get number of items in the ListView control ---

            mov     Index, 0

            fn      SendMessage, hControl, LVM_GETITEMCOUNT, 0, 0

            mov     ecx, eax

    ; --- Get item state ---

Get:        push    ecx

            fn      SendMessage, hControl, LVM_GETITEMSTATE, Index, LVIS_SELECTED

            cmp     eax, LVIS_SELECTED                                  ; Is selected?
            jne     Next

    ; --- Get item's name and extension ---

            mov     LVI.imask, LVIF_TEXT
            mov     eax, Index
            mov     LVI.iItem, eax
            mov     LVI.iSubItem, 0
            mov     LVI.state, 0
            mov     LVI.stateMask, 0
            lea     eax, FileName
            mov     LVI.pszText, eax
            mov     LVI.cchTextMax, sizeof FileName
            mov     LVI.iImage, 0
            mov     LVI.lParam, 0

            fn      SendMessage, hControl, LVM_GETITEM, 0, addr LVI

            mov     LVI.iSubItem, 1
            lea     eax, FileExt
            mov     LVI.pszText, eax
            mov     LVI.cchTextMax, sizeof FileExt

            fn      SendMessage, hControl, LVM_GETITEM, 0, addr LVI

    ; --- Get file handle ---

            fn      TrsCat, addr TrsName, addr FileName, addr FileExt   ; Concatenate name and extention

            mcall   OSI_VTBL, GetFileHandle, addr GUI_OSI, addr TrsName

            cmp     eax, -1                                             ; Error
            je      Error

            cmp     eax, NULL                                           ; Not found
            jne     @F

Error:      fn      MessageBox, hWnd, "OSI_GetFileHandle()", "Error", MB_ICONERROR+MB_OK
            jmp     Next

@@:         mov     hFPDE, eax

    ; --- Get the file size ---

            mcall   OSI_VTBL, GetFileSize, addr GUI_OSI, hFPDE

            or      eax, eax                                            ; Check whether the file is empty
            jz      Next

            mov     FileSize, eax

    ; --- Set file pointer ---

            mcall   OSI_VTBL, SetFilePointer, addr GUI_OSI, hFPDE, 0

            or      eax, eax
            jz      @F

            fn      MessageBox, hWnd, "OSI_SetFilePointer()", "Error", MB_ICONERROR+MB_OK
            jmp     Next

    ; --- Read entire file into buffer ---

@@:         mcall   OSI_VTBL, ReadFile, addr GUI_OSI, hFPDE, pBuffer, FileSize

            or      eax, eax
            jz      @F

            fn      MessageBox, hWnd, "OSI_ReadFile()", "Error", MB_ICONERROR+MB_OK
            jmp     Next

    ; --- Make Windows file name ---

@@:         fn      Trs2Win, addr TrsName, addr WinName                 ; Convert TRS-formatted name to Windows format (Name.Ext)

            fn      szCopy, addr Path, addr File                            ; Copy selected path to temporary buffer

            fn      szCatStr, addr File, addr WinName                       ; Append the Windows-formatted name to the path

    ; --- Create file and write buffer contents to disk ---

            fn      CreateFile, addr File, GENERIC_READ+GENERIC_WRITE, 0, NULL, CREATE_ALWAYS, 0, NULL
            mov     hFile, eax

            fn      WriteFile, hFile, pBuffer, FileSize, addr Bytes, NULL

            fn      CloseHandle, hFile

    ; --- Move to next ListView item ---

Next:       inc     Index
            pop     ecx
            dec     ecx
            jnz     Get

            fn      MessageBox, hWnd, "Operation completed.", "Export", MB_ICONINFORMATION+MB_OK

    ; --- Release allocated memory ---
            
FreeBuffer: fn      GlobalFree, pBuffer

Done:       ret

GUI_Export  endp
