;---------------------------------------------------------------------
; Graphical User Interface             Copyright (C) 2009 Miguel Dutra
;---------------------------------------------------------------------

GUI_Viewer          proto   :HWND, :UINT, :WPARAM, :LPARAM
GUI_Viewer_Cmd      proto   :HWND, :WPARAM, :LPARAM
GUI_Viewer_Init     proto   :HWND
GUI_Viewer_Dump     proto   :HWND

;---------------------------------------------------------------------
; GUI_Viewer                            Sector Viewer dialog procedure
;---------------------------------------------------------------------

GUI_Viewer  proc    hWnd:HWND, uMsg:UINT, wParam:WPARAM, lParam:LPARAM

    ; --- Handle the WM_INITDIALOG message ---
    
            cmp     uMsg, WM_INITDIALOG
            jne     @F

            fn      GUI_Viewer_Init, hWnd

            mov     eax, TRUE
            jmp     Done
                
    ; --- Handle the WM_CLOSE message ---
    
@@:         cmp     uMsg, WM_CLOSE
            jne     @F

            fn      EndDialog, hWnd, 0

            mov     eax, 0
            jmp     Done

    ; --- Handle the WM_COMMAND message ---

@@:         cmp     uMsg, WM_COMMAND
            jne     @F

            fn      GUI_Viewer_Cmd, hWnd, wParam, lParam
            jmp     Done

    ; --- Handle the WM_VSCROLL message ---

@@:         cmp     uMsg, WM_VSCROLL
            jne     @F

            fn      GUI_Viewer_Dump, hWnd

            mov     eax, 0
            jmp     Done

    ; --- Message not processed ---

@@:         mov     eax, FALSE
            
Done:       ret

GUI_Viewer  endp

;---------------------------------------------------------------------
; GUI_Viewer_Cmd                   Sector Viewer dialog initialization
;---------------------------------------------------------------------

GUI_Viewer_Cmd      proc    hWnd:HWND, wParam:WPARAM, lParam:LPARAM

            mov     eax, wParam                                         ; The high word contains the notification code
            shr     eax, 16
            cmp     eax, EN_CHANGE
            jne     @F

            fn      GUI_Viewer_Dump, hWnd

            mov     eax, 0
            jmp     Done

@@:         mov     eax, wParam                                         ; The low word contains the control id
            and     eax, 0FFFFh
            cmp     eax, IDC_EDITOR_CLOSE
            jne     @F

            fn      EndDialog, hWnd, 0

            mov     eax, 0
            jmp     Done

@@:         mov     eax, 1

Done:       ret

GUI_Viewer_Cmd      endp

;---------------------------------------------------------------------
; GUI_Viewer_Init                  Sector Viewer dialog initialization
;---------------------------------------------------------------------

GUI_Viewer_Init     proc    hWnd:HWND

            local   FirstTrack:DWORD
            local   LastTrack:DWORD
            local   FirstSide:DWORD
            local   LastSide:DWORD
            local   FirstSector:DWORD
            local   LastSector:DWORD
            local   String[8]:BYTE

            cmp     GUI_hFile, INVALID_HANDLE_VALUE
            je      Done

    ; --- Get disk information --- 

            mcall   VDI_VTBL, GetFirstTrack, addr GUI_VDI
            mov     FirstTrack, eax

            mcall   VDI_VTBL, GetLastTrack, addr GUI_VDI
            mov     LastTrack, eax

            mcall   VDI_VTBL, GetFirstSide, addr GUI_VDI
            mov     FirstSide, eax

            mcall   VDI_VTBL, GetLastSide, addr GUI_VDI
            mov     LastSide, eax

            mcall   VDI_VTBL, GetFirstSector, addr GUI_VDI
            mov     FirstSector, eax

            mcall   VDI_VTBL, GetLastSector, addr GUI_VDI
            mov     LastSector, eax

    ; --- Initialize buddy controls ---

            fn      wsprintf, addr String, "%u", FirstTrack             ; Track

            fn      GetDlgItem, hWnd, IDC_EDITOR_TRACK

            mov     edx, eax

            fn      SendMessage, edx, WM_SETTEXT, 0, addr String

            fn      wsprintf, addr String, "%u", FirstSide              ; Side

            fn      GetDlgItem, hWnd, IDC_EDITOR_SIDE

            mov     edx, eax

            fn      SendMessage, edx, WM_SETTEXT, 0, addr String

            fn      wsprintf, addr String, "%u", FirstSector            ; Sector

            fn      GetDlgItem, hWnd, IDC_EDITOR_SECTOR

            mov     edx, eax

            fn      SendMessage, edx, WM_SETTEXT, 0, addr String

    ; --- Set Up-Down control ranges ---

            fn      GetDlgItem, hWnd, IDC_EDITOR_SPIN1                  ; Track

            mov     edx, LastTrack
            shl     edx, 16
            add     edx, FirstTrack

            fn      SendMessage, eax, UDM_SETRANGE, 0, edx

            fn      GetDlgItem, hWnd, IDC_EDITOR_SPIN2                  ; Side

            mov     edx, LastSide
            shl     edx, 16
            add     edx, FirstSide

            fn      SendMessage, eax, UDM_SETRANGE, 0, edx

            fn      GetDlgItem, hWnd, IDC_EDITOR_SPIN3                  ; Sector

            mov     edx, LastSector
            shl     edx, 16
            add     edx, FirstSector

            fn      SendMessage, eax, UDM_SETRANGE, 0, edx

    ; --- Dump the initial sector ---

            fn      GUI_Viewer_Dump, hWnd

Done:       ret

GUI_Viewer_Init     endp

;---------------------------------------------------------------------
; GUI_Viewer_Dump                        Sector Viewer dumping routine
;---------------------------------------------------------------------

GUI_Viewer_Dump     proc    hWnd:HWND

            local   Text[16]:BYTE
            local   Buffer[1024]:BYTE
            local   DumpHex[((1024*3)+(1024/16*2))+1]:BYTE
            local   DumpAsc[(1024/16*(16+2))+1]:BYTE
            local   BufPtr:DWORD
            local   HexPtr:DWORD
            local   AscPtr:DWORD
            local   Track:BYTE
            local   Side:BYTE
            local   Sector:BYTE
            local   Count:BYTE

            cmp     GUI_hFile, INVALID_HANDLE_VALUE
            je      Done

            mov     Count, 1

            lea     eax, Buffer
            mov     BufPtr, eax

            lea     eax, DumpHex
            mov     HexPtr, eax

            lea     eax, DumpAsc
            mov     AscPtr, eax

            mov     ecx, sizeof DumpHex
@@:         mov     byte ptr DumpHex[ecx-1], 0
            loop    @B

            mov     ecx, sizeof DumpAsc
@@:         mov     byte ptr DumpAsc[ecx-1], 0
            loop    @B

    ; --- Retrieve the current buddy control values ---

            fn      GetDlgItem, hWnd, IDC_EDITOR_TRACK
            mov     edx, eax

            fn      SendMessage, edx, WM_GETTEXT, sizeof Text, addr Text
            fn      atodw, addr Text
            mov     Track, al

            fn      GetDlgItem, hWnd, IDC_EDITOR_SIDE
            mov     edx, eax

            fn      SendMessage, edx, WM_GETTEXT, sizeof Text, addr Text
            fn      atodw, addr Text
            mov     Side, al

            fn      GetDlgItem, hWnd, IDC_EDITOR_SECTOR
            mov     edx, eax

            fn      SendMessage, edx, WM_GETTEXT, sizeof Text, addr Text
            fn      atodw, addr Text
            mov     Sector, al

    ; --- Read the currently selected sector ---

            mcall   VDI_VTBL, ReadSector, addr GUI_VDI, addr Buffer, sizeof Buffer, Track, Side, Sector

            or      eax, eax
            jz      Dump

            mov     edx, eax

            fn      FormatMessage, FORMAT_MESSAGE_FROM_SYSTEM, NULL, edx, 0, addr DumpHex, sizeof DumpHex, NULL

            jmp     Assign

    ; --- Dump sector into strings ---

Dump:       push    ecx

            mov     eax, BufPtr
            movzx   eax, byte ptr [eax]

            fn      wsprintf, HexPtr, "%02X ", eax
            add     HexPtr, 3

            cmp     Count, 16
            jne     @F

            fn      wsprintf, HexPtr, "%c%c", 0Dh, 0Ah
            add     HexPtr, 2

@@:         mov     eax, BufPtr
            movzx   eax, byte ptr [eax]

            cmp     al, ' '
            jae     @F

            mov     al, '.'

@@:         fn      wsprintf, AscPtr, "%c", eax
            add     AscPtr, 1

            cmp     Count, 16
            jne     @F

            fn      wsprintf, AscPtr, "%c%c", 0Dh, 0Ah
            add     AscPtr, 2

            mov     Count, 0

@@:         inc     BufPtr
            inc     Count
            pop     ecx
            dec     ecx
            jnz     Dump

    ; --- Assign strings to controls ---

Assign:     fn      GetDlgItem, hWnd, IDC_EDITOR_HEX

            mov     edx, eax

            fn      SendMessage, edx, WM_SETTEXT, 0, addr DumpHex

            fn      GetDlgItem, hWnd, IDC_EDITOR_ASC

            mov     edx, eax

            fn      SendMessage, edx, WM_SETTEXT, 0, addr DumpAsc

Done:       ret

GUI_Viewer_Dump     endp

