;-----------------------------------------------------------
; OSI Module Checking Tool        Copyright (C) Miguel Dutra
;-----------------------------------------------------------

.386
.model flat, stdcall
option casemap :none

include \masm32\include\windows.inc
include \masm32\include\kernel32.inc
include \masm32\include\masm32.inc
include \masm32\macros\macros.asm

includelib \masm32\lib\kernel32.lib
includelib \masm32\lib\user32.lib
includelib \masm32\lib\masm32.lib

include ..\lib\macros.asm
include ..\lib\util.asm
include ..\VDI\VDI.asm
include ..\DMK\DMK.asm
include ..\JV3\JV3.asm
include ..\JV1\JV1.asm
include ..\OSI\OSI.asm
include ..\TRSDOS\TRSDOS.asm
include ..\MULTIDOS\MULTIDOS.asm

DiskTests   proto   :DWORD
List        proto   :DWORD
FileTests   proto   :DWORD
GetName     proto   :DWORD, :DWORD, :DWORD
GenName     proto   :DWORD, :DWORD, :DWORD
GetArg      proto

.code

;-----------------------------------------------------------
; Entry Point
;-----------------------------------------------------------

Main        proc

            local   argv[16]:DWORD
            local   argc:BYTE
            local   pVDI:DWORD
            local   pOSI:DWORD
            local   hFile:HANDLE

    ; --- Open the image file --- 

            fn      GetCommandLine
            
            fn      MakeArgList, addr argv, addr argc, eax
            
            fn      CreateFile, argv[1*DWORD], GENERIC_READ+GENERIC_WRITE, 0, NULL, OPEN_EXISTING, 0, NULL

            cmp     eax, INVALID_HANDLE_VALUE
            jne     @F

            fn      GetLastError
            fn      PrintError, eax
            jmp     Done

@@:         mov     hFile, eax

    ; --- Create and initialize a VDI object --- 
    
            fn      GlobalAlloc, GPTR, 256

            or      eax, eax
            jnz     @F
            
            fn      GetLastError
            fn      PrintError, eax
            jmp     Done
                     
@@:         mov     pVDI, eax

    ; --- Try to initialize the VDI interface as DMK ---

            fnc     StdOut, "DMK: "

            mov     eax, pVDI
                        
            mov     (CVDI PTR [eax]).VPTR, offset DMK

            mcall   VDI_VTBL, Load, pVDI, hFile, 0

            fn      PrintError, eax

            or      eax, eax
            jz      @F

    ; --- Try to initialize the VDI interface as JV3 ---

            fnc     StdOut, "JV3: "

            mov     eax, pVDI
                        
            mov     (CVDI PTR [eax]).VPTR, offset JV3

            mcall   VDI_VTBL, Load, pVDI, hFile, 0

            fn      PrintError, eax

            or      eax, eax
            jz      @F

    ; --- Try to initialize the VDI interface as JV1 ---

            fnc     StdOut, "JV1: "

            mov     eax, pVDI
                        
            mov     (CVDI PTR [eax]).VPTR, offset JV1

            mcall   VDI_VTBL, Load, pVDI, hFile, 0

            fn      PrintError, eax

            or      eax, eax
            jnz     DestroyVDI
            
    ; --- Create and initialize a OSI object --- 
    
@@:         fn      GlobalAlloc, GPTR, 256

            or      eax, eax
            jnz     @F
            
            fn      GetLastError
            fn      PrintError, eax
            jmp     DestroyVDI
                     
@@:         mov     pOSI, eax

            mov     (COSI PTR [eax]).VPTR, offset MULTIDOS              ; Patch

    ; --- Initialize the OSI interface --- 

            fnc     StdOut, "Load: "
            
            mcall   OSI_VTBL, Load, pOSI, pVDI, 0

            fn      PrintError, eax

            or      eax, eax
            jnz     DestroyOSI
            
    ; --- Execute testing procedures --- 

@@:         fn      DiskTests, pOSI

;            fn      PrintError, eax

            fn      List, pOSI

            fn      PrintError, eax

            fn      FileTests, pOSI

            fn      PrintError, eax

    ; --- Release the OSI interface --- 

UnloadOSI:  mcall   OSI_VTBL, Unload, pOSI
            
            or      eax, eax
            jz      DestroyOSI
            
            fn      PrintError, eax

    ; --- Delete the OSI object --- 

DestroyOSI: fn      GlobalFree, pOSI

            or      eax, eax
            jz      UnloadVDI

            fn      GetLastError
            fn      PrintError, eax

    ; --- Release the VDI interface --- 

UnloadVDI:  mcall   VDI_VTBL, Unload, pVDI
            
            or      eax, eax
            jz      DestroyVDI
            
            fn      PrintError, eax

    ; --- Delete the VDI object --- 

DestroyVDI: fn      GlobalFree, pVDI

            or      eax, eax
            jz      Close

            fn      GetLastError
            fn      PrintError, eax

    ; --- Close the file handle --- 

Close:      fn      CloseHandle, hFile

            or      eax, eax
            jnz     Done

            fn      GetLastError
            fn      PrintError, eax

Done:       fn      ExitProcess, 0

Main        endp

;-----------------------------------------------------------
; Perform disk operations
;-----------------------------------------------------------

DiskTests   proc    pOSI:DWORD

            local   abDiskName[9]:BYTE
            local   abDiskDate[9]:BYTE
            local   dwDosVersion:DWORD
            local   abString[80]:BYTE

            mcall   OSI_VTBL, GetDosVersion, pOSI
            mov     dwDosVersion, eax

            mcall   OSI_VTBL, GetDiskName, pOSI, addr abDiskName, sizeof abDiskName
            mov     abDiskName[8], 0

            mcall   OSI_VTBL, GetDiskDate, pOSI, addr abDiskDate, sizeof abDiskDate
            mov     abDiskDate[8], 0

            fnc     wsprintf, addr abString, "DOS Version: %02X, Disk Name: %s, Disk Date: %s\n", dwDosVersion, addr abDiskName, addr abDiskDate
            fn      StdOut, addr abString

            ret

DiskTests   endp

;-----------------------------------------------------------
; List files
;-----------------------------------------------------------

List        proc    pOSI:DWORD

            local   bFindFlag:BYTE
            local   pFPDE:DWORD
            local   abName[9]:BYTE
            local   abExt[4]:BYTE
            local   dwSize:DWORD
            local   dwYear:DWORD
            local   dwMonth:DWORD
            local   dwDay:DWORD
            local   dwAttr:DWORD
            local   abString[80]:BYTE

            mov     bFindFlag, OSI_FINDFIRST

Next:       mcall   OSI_VTBL, Find, pOSI, bFindFlag

            or      eax, eax
            jnz     @F

            mcall   OSI_VTBL, GetLastError, pOSI
            jmp     Done

@@:         mov     pFPDE, eax

            mcall   OSI_VTBL, GetFileName, pOSI, pFPDE, addr abName, sizeof abName
            mov     abName[8], 0

            mcall   OSI_VTBL, GetFileExt, pOSI, pFPDE, addr abExt, sizeof abExt
            mov     abExt[3], 0

            mcall   OSI_VTBL, GetFileSize, pOSI, pFPDE
            mov     dwSize, eax

            mcall   OSI_VTBL, GetFileDate, pOSI, pFPDE

            movzx   edx, al
            mov     dwDay, edx

            shr     eax, 8
            movzx   edx, al
            mov     dwMonth, edx

            shr     eax, 8
            movzx   edx, ax
            mov     dwYear, edx

            mcall   OSI_VTBL, GetFileAttr, pOSI, pFPDE
            mov     dwAttr, eax

            fnc     wsprintf, addr abString, "%s%s %6i %04i-%02i-%02i %04X\n", addr abName, addr abExt, dwSize, dwYear, dwMonth, dwDay, dwAttr
            fn      StdOut, addr abString

            mov     bFindFlag, OSI_FINDNEXT
            jmp     Next

Done:       ret

List        endp

;-----------------------------------------------------------
; Perform file operations
;-----------------------------------------------------------

FileTests   proc    pOSI:DWORD

            local   abFile[12]:BYTE
            local   abTemp[12]:BYTE
            local   String[80]:BYTE
            local   pSource:DWORD
            local   pTarget:DWORD
            local   pBuffer:DWORD
            local   dwSize:DWORD
            local   dwDate:DWORD
            local   dwAttr:DWORD
            local   dwSourceSum:DWORD
            local   dwTargetSum:DWORD
            local   abString[80]:BYTE

            mov     abFile[sizeof abFile - 1], 0
            mov     abTemp[sizeof abTemp - 1], 0

    ; --- Get file name ---
    
            fn      GetName, pOSI, addr abFile, 11

            fnc     wsprintf, addr abString, "Reading ""%s""...\n", addr abFile
            fn      StdOut, addr abString

    ; --- Open file ---
    
            mcall   OSI_VTBL, Open, pOSI, addr abFile

            or      eax, eax
            jnz     @F

            mcall   OSI_VTBL, GetLastError, pOSI
            jmp     Done

@@:         mov     pSource, eax

    ; --- Get file information ---
    
            mcall   OSI_VTBL, GetFileSize, pOSI, pSource
            
            mov     dwSize, eax

            mcall   OSI_VTBL, GetFileDate, pOSI, pSource
            
            mov     dwDate, eax

            mcall   OSI_VTBL, GetFileAttr, pOSI, pSource
            
            mov     dwAttr, eax

    ; --- Allocate memory ---
    
            fn      GlobalAlloc, GPTR, dwSize

            or      eax, eax
            jnz     @F

            fn      GetLastError
            jmp     Done

@@:         mov     pBuffer, eax

    ; --- Read from file ---

            mcall   OSI_VTBL, Seek, pOSI, pSource, 0

            or      eax, eax
            jnz     Release
            
            mcall   OSI_VTBL, Read, pOSI, pSource, pBuffer, dwSize

            or      eax, eax
            jnz     Release

            fn      Dump, pBuffer, dwSize

    ; --- Create new file ---

            fn      GenName, addr abFile, addr abTemp, 11

            fnc     wsprintf, addr abString, "Creating ""%s""...\n", addr abTemp
            fn      StdOut, addr abString

            mcall   OSI_VTBL, Create, pOSI, addr abTemp, dwSize

            or      eax, eax
            jnz     @F

            mcall   OSI_VTBL, GetLastError, pOSI
            jmp     Release

@@:         mov     pTarget, eax

    ; --- Set file attributes ---

            mcall   OSI_VTBL, SetFileDate, pOSI, pTarget, dwDate

            or      eax, eax
            jnz     Release
            
            mcall   OSI_VTBL, SetFileAttr, pOSI, pTarget, dwAttr

            or      eax, eax
            jnz     Release

    ; --- Write to file ---

            fnc     wsprintf, addr abString, "Copying ""%s"" to ""%s""...\n", addr abFile, addr abTemp
            fn      StdOut, addr abString

            mcall   OSI_VTBL, Seek, pOSI, pTarget, 0

            or      eax, eax
            jnz     Release
            
            mcall   OSI_VTBL, Write, pOSI, pTarget, pBuffer, dwSize

            or      eax, eax
            jnz     Release
            
    ; --- Delete original file ---

            fnc     wsprintf, addr abString, "Deleting ""%s""...\n", addr abFile
            fn      StdOut, addr abString

            mcall   OSI_VTBL, Delete, pOSI, pSource

            or      eax, eax
            jnz     Release

    ; --- Rename temporary file ---

            fnc     wsprintf, addr abString, "Renaming ""%s"" to ""%s""...\n", addr abTemp, addr abFile
            fn      StdOut, addr abString

            mcall   OSI_VTBL, Rename, pOSI, pTarget, addr abFile

            or      eax, eax
            jnz     Release

    ; --- Release allocated memory ---

Release:    push    eax
            fn      GlobalFree, pBuffer
            pop     eax

Done:       ret

FileTests   endp

;-----------------------------------------------------------
; Return the last file name in the disk
;-----------------------------------------------------------

GetName     proc    pOSI:DWORD, pStr:DWORD, Len:DWORD

            local   pFPDE:DWORD
            local   FindFlag:BYTE

            cmp     Len, 11
            jb      Done
            
            mov     FindFlag, OSI_FINDFIRST

Next:       mcall   OSI_VTBL, Find, pOSI, FindFlag

            or      eax, eax
            jz      Done

            mov     pFPDE, eax

            mcall   OSI_VTBL, GetFileName, pOSI, pFPDE, pStr, 8
            
            mov     edx, pStr
            add     edx, 8

            mcall   OSI_VTBL, GetFileExt, pOSI, pFPDE, edx, 3
            
            mov     FindFlag, OSI_FINDNEXT
            jmp     Next

Done:       ret
            
GetName     endp

;-----------------------------------------------------------
; Generate a new file name from a previous one
;-----------------------------------------------------------

GenName     proc    pSource:DWORD, pTarget:DWORD, dwLen:DWORD

            push    esi
            push    edi
            
            mov     esi, pSource
            mov     edi, pTarget
            mov     ecx, dwLen

Gen:        lodsb

            cmp     al, ' '
            je      Sto

            cmp     al, '9'
            jne     @F
            
            mov     al, '0'
            jmp     Sto
            
@@:         cmp     al, 'Z'
            jne     @F

            mov     al, 'A'
            jmp     Sto
            
@@:         inc     al
            
Sto:        stosb
            loop    Gen

            pop     edi
            pop     esi

            ret
            
GenName     endp

;-----------------------------------------------------------
; Return pointer to first command-line argument
;-----------------------------------------------------------

GetArg      proc

            fn      GetCommandLine

            mov     dl, 1                                               ; Looking for blank (0:No, 1:Yes)
                   
Next:       cmp     byte ptr [eax], ' '                                 ; Is blank?
            sete    dh
            
            cmp     dh, dl                                              ; Match our search?
            jne     @F

            cmp     dl, 1                                               ; Have we gone through "yes" and "no"?
            jne     Done

            mov     dl, 0                                               ; Repeat looking for a non-blank

@@:         cmp     byte ptr [eax], 0                                   ; Have we reached the end of line?
            je      Done
    
            inc     eax                                                 ; Point to the next char
            jmp     Next
    
Done:       ret

GetArg      endp

end Main

