;---------------------------------------------------------------------
; RAPIDOS Operating System Interface        Copyright (C) Miguel Dutra
;---------------------------------------------------------------------

.data

;---------------------------------------------------------------------
; RAPIDOS Virtual Table
;---------------------------------------------------------------------

RAPIDOS_VTBL    struct                                                  ; Declare the structure and initializers
    Load                    DWORD       RD_Load                         ; OSI-derived functions ; Patch
    Unload                  DWORD       TD_Unload
    Find                    DWORD       TD_Find
    Open                    DWORD       TD_Open
    Create                  DWORD       TD_Create
    Seek                    DWORD       TD_Seek
    Read                    DWORD       TD_Read
    Write                   DWORD       TD_Write
    Rename                  DWORD       TD_Rename
    Delete                  DWORD       TD_Delete
    GetLastError            DWORD       TD_GetLastError
    GetDosVersion           DWORD       TD_GetDosVersion
    SetDosVersion           DWORD       TD_SetDosVersion
    GetDiskName             DWORD       TD_GetDiskName
    SetDiskName             DWORD       TD_SetDiskName
    GetDiskDate             DWORD       TD_GetDiskDate
    SetDiskDate             DWORD       TD_SetDiskDate
    GetFilePos              DWORD       TD_GetFilePos
    GetFileName             DWORD       TD_GetFileName
    GetFileExt              DWORD       TD_GetFileExt
    GetFileSize             DWORD       TD_GetFileSize
    GetFileDate             DWORD       TD_GetFileDate
    SetFileDate             DWORD       TD_SetFileDate
    GetFileAttr             DWORD       TD_GetFileAttr
    SetFileAttr             DWORD       TD_SetFileAttr
    DirRW                   DWORD       TD_DirRW                        ; TRSDOS-specific functions
    GetFDE                  DWORD       TD_GetFDE
    GetExtPtr               DWORD       TD_GetExtPtr
    AllocGAT                DWORD       TD_AllocGAT
    FreeGAT                 DWORD       TD_FreeGAT
    ScanHIT                 DWORD       TD_ScanHIT
    SetHIT                  DWORD       TD_SetHIT
    DEC2FDE                 DWORD       TD_DEC2FDE
    FDE2DEC                 DWORD       TD_FDE2DEC
    Hash                    DWORD       TD_Hash
    Sector2CHS              DWORD       TD_Sector2CHS
RAPIDOS_VTBL    ends

RAPIDOS                     RAPIDOS_VTBL    <>                          ; Define a variable of that type

.code

;---------------------------------------------------------------------
; RD_Load                                         Initialize interface
;---------------------------------------------------------------------
; pRD:      Pointer to the calling RAPIDOS object
; pVDI:     Pointer to an initialized VDI object
;---------------------------------------------------------------------
; EAX:      Error code
;---------------------------------------------------------------------

RD_Load     proc    pRD:DWORD, pVDI:DWORD

            local   abBoot[256]:BYTE
            local   pDir:DWORD
            local   bDirTrack:BYTE
            local   bDirSectors:BYTE
            local   bFirstTrack:BYTE
            local   bFirstSide:BYTE
            local   bFirstSector:BYTE
            local   bSides:BYTE
            local   bSectorsPerTrack:BYTE
            local   bGranulesPerTrack:BYTE
            local   bGranulesPerCylinder:BYTE
            local   bSectorsPerGranule:BYTE
            local   bDensity:BYTE

    ; --- Get disk information ---

            mcall   VDI_VTBL, GetFirstTrack, pVDI
            mov     bFirstTrack, al

            mcall   VDI_VTBL, GetFirstSide, pVDI
            mov     bFirstSide, al

            mcall   VDI_VTBL, GetSides, pVDI
            mov     bSides, al

            mcall   VDI_VTBL, GetDensity, pVDI
            mov     bDensity, al

            mcall   VDI_VTBL, GetFirstSector, pVDI                      ; Patch
            mov     bFirstSector, al

    ; --- Normalize Sectors Per Track ---

            mcall   VDI_VTBL, GetSPT, pVDI                              ; Get Sectors Per Track

            cmp     al, 26                                              ; 8" DD                                              
            jb      @F

            mov     al, 26
            jmp     SetSPT
            
@@:         cmp     al, 18                                              ; 5 1/4" DD
            jb      @F

            mov     al, 18
            jmp     SetSPT
            
@@:         cmp     al, 15                                              ; 8" SD
            jb      @F

            mov     al, 15
            jmp     SetSPT
            
@@:         mov     al, 10                                              ; 5 1/4" SD

SetSPT:     mov     bSectorsPerTrack, al

    ; --- Load the boot sector to find out the directory cylinder ---

            mov     al, bFirstSector                                    ; Patch
            inc     al
            
            mcall   VDI_VTBL, Read, pVDI, addr abBoot, sizeof abBoot, bFirstTrack, bFirstSide, al   ; Patch
            
            or      eax, eax
            jnz     Done

            cmp     abBoot[1], 0FEh                                     ; Kind of integrity check
            jne     Error

            mov     al, abBoot[2]                                       ; This byte contains the directory cylinder
            mov     bDirTrack, al
            
    ; --- Compute the number of directory sectors ---

            mov     al, bSectorsPerTrack                                ; Directory Sectors = Sectors per Track * Sides
            mul     bSides

            cmp     al, 32+2                                            ; Max sectors limited by HIT: 256 hashes / 8 FDEs per sector = 32 sectors (+2 for GAT and HIT)
            jbe     @F

            mov     al, 32+2                                            ; Set top limit

@@:         mov     bDirSectors, al
            
    ; --- Allocate memory for the entire directory ---

            shl     eax, 8                                              ; Buffer Size = DirSectors * 256 bytes per sector (TRSDOS standard)

            fn      GlobalAlloc, GPTR, eax

            or      eax, eax
            jnz     @F

            fn      GetLastError
            jmp     Done

@@:         mov     pDir, eax                                           ; Save pointer to Directory buffer

    ; --- Update member class variables ---
    
            mov     edx, pRD
            
            mov     eax, pVDI                                           ; Update pointer to Virtual Disk Interface
            mov     (CTRSDOS PTR [edx]).pVDI, eax

            mov     eax, pDir                                           ; Update pointer to Directory buffer
            mov     (CTRSDOS PTR [edx]).pDir, eax

            mov     al, bDirTrack                                       ; Update number of the directory cylinder
            mov     (CTRSDOS PTR [edx]).bDirTrack, al

            mov     al, bDirSectors                                     ; Update number of directory sectors
            mov     (CTRSDOS PTR [edx]).bDirSectors, al

            mov     al, bFirstTrack                                     ; Update number of the first track on disk
            mov     (CTRSDOS PTR [edx]).bFirstTrack, al

            mov     al, bFirstSide                                      ; Update number of the first disk side
            mov     (CTRSDOS PTR [edx]).bFirstSide, al

            mov     al, bFirstSector                                    ; Update number of the first sector on track
            mov     (CTRSDOS PTR [edx]).bFirstSector, al

            mov     al, bSides                                          ; Update number of disk sides
            mov     (CTRSDOS PTR [edx]).bSides, al

            mov     al, bSectorsPerTrack                                ; Update number of Sectors per Track
            mov     (CTRSDOS PTR [edx]).bSectorsPerTrack, al

    ; --- Load directory into memory ---

            mcall   TRSDOS_VTBL, DirRW, pRD, TD_DIR_READ, 0
            
            or      eax, eax
            jnz     Done

    ; --- Compute and validate the disk's granule units ---

            mov     edx, pDir                                           ; Pointer to the first directory sector (GAT)

            mov     cl, (TD_GAT PTR [edx]).bDiskConfig                  ; Find the number of granules per track
            and     cl, TD_GAT_Granules
            inc     cl
            mov     bGranulesPerTrack, cl

            cmp     cl, 2                                               ; Granules Per Track must be between 2 and 8    ; Patch
            jb      Error

;            cmp     cl, 8                                              ; Por que est comentado?
;            ja      Error
            
            mov     al, bSides                                          ; Find the number of granules per cylinder
            mul     cl
            mov     bGranulesPerCylinder, al

            mov     al, bSectorsPerTrack                                ; Compute the number of sectors per granule
            div     cl                                                  ; Patch
            mov     bSectorsPerGranule, al

            or      ah, ah                                              ; There must be no remainder after last division
            jnz     Error

            cmp     al, 5                                               ; Sectors per Granule for 5.25" SDEN disks
            je      Single
    
            cmp     al, 6                                               ; Sectors per Granule for 5.25" DDEN disks
            je      Double
    
            cmp     al, 8                                               ; Sectors per Granule for 8" SDEN disks
            je      Single
    
            cmp     al, 10                                              ; Sectors per Granule for 8" DDEN disks
            jne     Error
    
    ; --- Check if the SPG conforms with the disk density ---

Double:     cmp     bDensity, 2                                         ; Check if the disk is double or mixed density
            je      @F

            cmp     bDensity, 1
            je      @F
            jmp     Error

Single:     cmp     bDensity, 2                                         ; Check if the disk is single or mixed density
            je      @F

            cmp     bDensity, 0
            je      @F
            jmp     Error

    ; --- Update member class variables ---
    
@@:         mov     edx, pRD
            
            mov     al, bGranulesPerTrack                               ; Update number of Granules Per Track
            mov     (CTRSDOS PTR [edx]).bGranulesPerTrack, al

            mov     al, bGranulesPerCylinder                            ; Update number of Granules Per Cylinder
            mov     (CTRSDOS PTR [edx]).bGranulesPerCylinder, al

            mov     al, bSectorsPerGranule                              ; Update number of Sectors Per Granule
            mov     (CTRSDOS PTR [edx]).bSectorsPerGranule, al

            mov     eax, NO_ERROR                                       ; The operation completed successfully.
            jmp     Done

Error:      mov     eax, ERROR_NOT_DOS_DISK                             ; The specified disk or diskette cannot be accessed.

Done:       ret

RD_Load     endp

