;-----------------------------------------------------------
; VDI Module Checking Tool        Copyright (C) Miguel Dutra
;-----------------------------------------------------------

.386
.model flat, stdcall
option casemap :none

include \masm32\include\windows.inc
include \masm32\include\kernel32.inc
include \masm32\include\user32.inc
include \masm32\include\masm32.inc
include \masm32\macros\macros.asm

includelib \masm32\lib\kernel32.lib
includelib \masm32\lib\user32.lib
includelib \masm32\lib\masm32.lib

include ..\macros\macros.asm
include VDI.asm

.code

Dump        proto   :DWORD
PrintSector proto   :DWORD, :DWORD
GetArg      proto
PrintError  proto   :DWORD

;-----------------------------------------------------------
; Entry Point
;-----------------------------------------------------------

Main        proc

            local   pVDI:DWORD
            local   hFile:HANDLE
            local   Buffer[256]:BYTE
            local   Tracks:BYTE
            local   Sides:BYTE
            local   SPT:BYTE
            local   Density:BYTE
            local   FirstTrack:BYTE
            local   LastTrack:BYTE
            local   FirstSide:BYTE
            local   LastSide:BYTE
            local   FirstSector:BYTE
            local   LastSector:BYTE

    ; --- Open the image file --- 
    
            fnc     StdOut, "Opening file...\n"

            fn      GetArg

            fn      CreateFile, eax, GENERIC_READ+GENERIC_WRITE, 0, NULL, OPEN_EXISTING, 0, NULL

            cmp     eax, INVALID_HANDLE_VALUE
            jne     @F

            fn      GetLastError
            fn      PrintError, eax
            jmp     Done

@@:         mov     hFile, eax

    ; --- Create and initialize the VDI object --- 
    
            fnc     StdOut, "Creating VDI object...\n"

            fn      GlobalAlloc, GPTR, 256

            or      eax, eax
            jnz     @F
            
            fn      GetLastError
            fn      PrintError, eax
            jmp     Done
                     
@@:         mov     pVDI, eax

            mov     (CVDI PTR [eax]).VPTR, offset VDI

    ; --- Initialize the interface --- 

            fnc     StdOut, "Loading interface...\n"
            
            mcall   VDI_VTBL, Load, pVDI, hFile, 0

            or      eax, eax
            jz      @F
            
            fn      PrintError, eax
            jmp     Destroy

    ; --- Retrieve the disk geometry --- 

@@:         fnc     StdOut, "Retrieving disk geometry...\n"
            
            mcall   VDI_VTBL, GetLastSector, pVDI
            mov     LastSector, al
            push    eax

            mcall   VDI_VTBL, GetFirstSector, pVDI
            mov     FirstSector, al
            push    eax
            
            mcall   VDI_VTBL, GetLastSide, pVDI
            mov     LastSide, al
            push    eax
            
            mcall   VDI_VTBL, GetFirstSide, pVDI
            mov     FirstSide, al
            push    eax
            
            mcall   VDI_VTBL, GetLastTrack, pVDI
            mov     LastTrack, al
            push    eax
            
            mcall   VDI_VTBL, GetFirstTrack, pVDI
            mov     FirstTrack, al
            push    eax
            
            mcall   VDI_VTBL, GetDensity, pVDI
            mov     Density, al
            push    eax
            
            mcall   VDI_VTBL, GetSPT, pVDI
            mov     SPT, al
            push    eax
            
            mcall   VDI_VTBL, GetSides, pVDI
            mov     Sides, al
            push    eax
            
            mcall   VDI_VTBL, GetTracks, pVDI
            mov     Tracks, al
            push    eax
            
            fn      wsprintf, addr Buffer, "T=%d S=%d SPT=%d D=%d FT=%d LT=%d FS=%d LS=%d Fs=%d Ls=%d"
            fn      StdOut, addr Buffer
            add     esp, 4*10
            
    ; --- Read last disk sector --- 

            fnc     StdOut, "Reading last sector...\n"
            
            mcall   VDI_VTBL, Read, pVDI, addr Buffer, sizeof Buffer, LastTrack, LastSide, LastSector

            or      eax, eax
            jz      @F
            
            fn      PrintError, eax
            jmp     Destroy

    ; --- Write last disk sector --- 

@@:         fnc     StdOut, "Writing last sector...\n"

            mcall   VDI_VTBL, Write, pVDI, addr Buffer, sizeof Buffer, LastTrack, LastSide, LastSector

            or      eax, eax
            jz      @F
            
            fn      PrintError, eax
            jmp     Destroy

    ; --- Dump all disk sectors --- 

@@:         fnc     StdOut, "Dumping disk sectors...\n"

            fn      Dump, pVDI

    ; --- Release the interface --- 

            fnc     StdOut, "Unloading interface...\n"

            mcall   VDI_VTBL, Unload, pVDI

            or      eax, eax
            jz      Destroy
            
            fn      PrintError, eax

    ; --- Delete the VDI object --- 

Destroy:    fnc     StdOut, "Destroying VDI object...\n"
            
            fn      GlobalFree, pVDI

Done:       fn      ExitProcess, 0

Main        endp

;-----------------------------------------------------------
; Dump disk sectors
;-----------------------------------------------------------

Dump        proc    pVDI:DWORD

            local   Buffer[1024]:BYTE
            local   String[64]:BYTE
            local   LastTrack:BYTE
            local   FirstSector:BYTE
            local   LastSector:BYTE
            local   FirstSide:BYTE
            local   LastSide:BYTE
            local   Track:BYTE
            local   Sector:BYTE
            local   Side:BYTE

    ; --- Retrieve the disk geometry ---

            mcall   VDI_VTBL, GetFirstTrack, pVDI
            mov     Track, al

            mcall   VDI_VTBL, GetLastTrack, pVDI
            mov     LastTrack, al

            mcall   VDI_VTBL, GetFirstSector, pVDI
            mov     FirstSector, al
            mov     Sector, 0                                           ; Always start from 0 (some disks have first track formatted differently)

            mcall   VDI_VTBL, GetLastSector, pVDI
            mov     LastSector, al

            mcall   VDI_VTBL, GetFirstSide, pVDI
            mov     FirstSide, al
            mov     Side, al

            mcall   VDI_VTBL, GetLastSide, pVDI
            mov     LastSide, al

    ; --- Read and dump the sector ---

DumpSector: mcall   VDI_VTBL, Read, pVDI, addr Buffer, sizeof Buffer, Track, Side, Sector    ; Read sector at first track, first side, sector 0

            or      eax, eax                                            ; On error, skip the sector
            jnz     Next

            push    ecx

            movzx   edx, Track
            movzx   ecx, Side
            movzx   eax, Sector
            
            fn      wsprintf, addr String, "%c%c[%02u:%u:%02u]%c%c", 0Dh, 0Ah, edx, ecx, eax, 0Dh, 0Ah
    
            fn      StdOut, addr String                                 ; Print the sector address (CC:H:SS)

            pop     ecx

            fn      PrintSector, addr Buffer, ecx                       ; Print sector data

    ; --- Advance to next Sector, Side, Track ---

Next:       inc     Sector                                              ; Increment sector
            mov     al, Sector
            cmp     al, LastSector
            jbe     DumpSector
        
            mov     al, FirstSector                                     ; Reset sector count
            mov     Sector, al

            inc     Side                                                ; Increment side
            mov     al, Side
            cmp     al, LastSide
            jbe     DumpSector

            mov     al, FirstSide                                       ; Reset side count
            mov     Side, al

            inc     Track                                               ; Increment track
            mov     al, Track
            cmp     al, LastTrack
            jbe     DumpSector

Done:       ret
    
Dump        endp

;-----------------------------------------------------------
; Print sector data in both HEX and ASCII
;-----------------------------------------------------------

PrintSector proc    pBuffer:DWORD, Len:DWORD

            local   Hex[48+1]:BYTE
            local   ASC[16+1]:BYTE
            local   HexPtr:DWORD
            local   AscPtr:DWORD

NewLine:    lea     eax, Hex
            mov     HexPtr, eax

            lea     eax, ASC
            mov     AscPtr, eax

GetChar:    mov     edx, pBuffer
            movzx   eax, byte ptr [edx]

            push    eax
            fn      wsprintf, HexPtr, "%02X ", eax
            pop     eax

            cmp     al, ' '
            jae     @F

            mov     al, '.'

@@:         fn      wsprintf, AscPtr, "%c", al

            add     HexPtr, 3
            inc     AscPtr
            inc     pBuffer
            dec     Len

            lea     eax, ASC[16]
            cmp     AscPtr, eax
            jb      GetChar

            fn      StdOut, addr Hex
            fn      StdOut, addr ASC
            fnc     StdOut, "\n"

            cmp     Len, 0
            ja      NewLine

Done:       ret

PrintSector endp

;-----------------------------------------------------------
; Return pointer to first command-line argument
;-----------------------------------------------------------

GetArg      proc

            fn      GetCommandLine

            mov     dl, 1                                               ; Looking for blank (0:No, 1:Yes)
                   
Next:       cmp     byte ptr [eax], ' '                                 ; Is blank?
            sete    dh
            
            cmp     dh, dl                                              ; Match our search?
            jne     @F

            cmp     dl, 1                                               ; Have we gone through "yes" and "no"?
            jne     Done

            mov     dl, 0                                               ; Repeat looking for a non-blank

@@:         cmp     byte ptr [eax], 0                                   ; Have we reached the end of line?
            je      Done
    
            inc     eax                                                 ; Point to the next char
            jmp     Next
    
Done:       ret

GetArg      endp

;-----------------------------------------------------------
; Print formatted system error messages
;-----------------------------------------------------------

PrintError  proc    Error:DWORD

            local   pMessage:PTR

            fn      FormatMessage, FORMAT_MESSAGE_FROM_SYSTEM or FORMAT_MESSAGE_ALLOCATE_BUFFER, NULL, Error, 0, addr pMessage, 0, NULL

            or      eax ,eax
            jz      @F

            fn      StdOut, pMessage

@@:         fn      LocalFree, pMessage

            mov     eax, Error
            
            ret

PrintError  endp

end Main

